# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from __future__ import print_function
import logging
import os
import sys

SSH = 0
SCP = 1
TAKTUK = 2
CHAINPUT = 3

def checktty(f):
    try:
        if (get_ipython().__class__.__name__ == 'ZMQInteractiveShell' #@UndefinedVariable
            and f.__class__.__module__ == 'IPython.zmq.iostream'
            and f.__class__.__name__ == 'OutStream'):
            return True
    except NameError:
        pass
    if hasattr(f, 'isatty'):
        return f.isatty()
    if hasattr(f, 'fileno'):
        return os.isatty(f.fileno())
    return False

IODEBUG = 3
FDEBUG = 5
TRACE = 12
DETAIL = 15

# _STARTOF_ configuration
configuration = {
    'log_level': logging.INFO,
    'remote_tool': SSH,
    'fileput_tool': SCP,
    'fileget_tool': SCP,
    'compact_output_threshold': 4096,
    'kill_timeout': 60,
    'intr_period': 1,
    'port_range': (25500, 26700),
    'kill_childs_at_end': True,
    'color_mode': checktty(sys.stdout)
                  and checktty(sys.stderr),
    'color_styles': {
        'log_header': ('yellow',),
        'object_repr': ('blue',),
        'emph': ('cyan',),
        'report_warn': ('magenta',),
        'report_error': ('red', 'bold'),
        'command': ('blue', 'bold'),
        'host': ('magenta', 'bold'),
        'user1': ('green', 'bold'),
        'user2': ('yellow', 'bold'),
        'user3': ('cyan', 'bold'),
        IODEBUG: ('green', 'reverse'),
        FDEBUG: ('green', 'reverse'),
        logging.DEBUG: ('green',),
        TRACE: ('green', 'bold',),
        DETAIL: ('magenta', 'bold'),
        logging.INFO: ('magenta',),
        logging.WARNING: ('cyan',),
        logging.ERROR: ('red',),
        logging.CRITICAL: ('yellow', 'on_red')
        },
    }
# _ENDOF_ configuration
"""Global execo configuration parameters.

- ``log_level``: the log level (see module `logging`)

- ``remote_tool``: default tool to use when instanciating remote
  processes. Can be `execo.config.SSH` or `execo.config.TAKTUK`

- ``fileput_tool``: default tool to use to put files remotely. Can be
  `execo.config.SCP` or `execo.config.TAKTUK`

- ``fileget_tool``: default tool to use to get remote files. Can be
  `execo.config.SCP` or `execo.config.TAKTUK`

- ``compact_output_threshold``: only beginning and end of stdout /
  stderr are displayed by `execo.process.ProcessBase.dump` when their
  size is greater than this threshold. 0 for no threshold

- ``kill_timeout``: number of seconds to wait after a clean SIGTERM
  kill before assuming that the process is not responsive and killing
  it with SIGKILL

- ``intr_period``: number of seconds between periodic check interrupt,
  for correct handling of ctrl-c

- ``port_range``: a tuple (start port, end port) of ports to use for
  the function ``execo.utils.get_port``. As all python ranges, start
  is inclusive, end is exclusive.

- ``kill_childs_at_end``: Whether to try sending SIGTERM to all
  subprocesses started through execo when the script
  terminates. Warnings:

  - this config option must be set at execo import time, changing it
    later will be ignored

  - SIGTERM is only sent to all childs or subchilds which did not try
    to daemonize by changing their process group.

- ``color_mode``: whether to colorize output (with ansi escape
  sequences)

- ``color_styles``: mapping of identifiers to iterables of ansi
  attributes identifiers (see `execo.log._ansi_styles`)
"""

def make_default_connection_params():
# _STARTOF_ default_connection_params
    default_connection_params = {
        'user':        None,
        'keyfile':     None,
        'port':        None,
        'ssh':         'ssh',
        'scp':         'scp',
        'taktuk':      'taktuk',
        'ssh_options': ( '-tt',
                         '-o', 'BatchMode=yes',
                         '-o', 'PasswordAuthentication=no',
                         '-o', 'StrictHostKeyChecking=no',
                         '-o', 'UserKnownHostsFile=/dev/null',
                         '-o', 'ConnectTimeout=20' ),
        'scp_options': ( '-o', 'BatchMode=yes',
                         '-o', 'PasswordAuthentication=no',
                         '-o', 'StrictHostKeyChecking=no',
                         '-o', 'UserKnownHostsFile=/dev/null',
                         '-o', 'ConnectTimeout=20',
                         '-rp' ),
        'taktuk_options': ( '-s', ),
        'taktuk_connector': 'ssh',
        'taktuk_connector_options': ( '-o', 'BatchMode=yes',
                                      '-o', 'PasswordAuthentication=no',
                                      '-o', 'StrictHostKeyChecking=no',
                                      '-o', 'UserKnownHostsFile=/dev/null',
                                      '-o', 'ConnectTimeout=20'),
        'chainput_nc_client_timeout': 5,
        'chainput_nc_server_timeout': 30,
        'nc': '/bin/nc.traditional -v -v',
        'chainput_host_retry': 5,
        'chainput_chain_retry': 10,
        'chainput_try_delay': 1,
        'forwarding_timeout': 25,
        'pty': False,
        'host_rewrite_func': None
        }
# _ENDOF_ default_connection_params
    return default_connection_params

default_connection_params = make_default_connection_params()
"""Default connection params for ``ssh``/``scp``/``taktuk`` connections.

- ``user``: the user to connect with.

- ``keyfile``: the keyfile to connect with.

- ``port``: the port to connect to.

- ``ssh``: the ssh or ssh-like command.

- ``scp``: the scp or scp-like command.

- ``taktuk``: the taktuk command.

- ``ssh_options``: tuple of options passed to ssh.

- ``scp_options``: tuple of options passed to scp.

- ``taktuk_options``: tuple of options passed to taktuk.

- ``taktuk_connector``: the ssh-like connector command for taktuk.

- ``taktuk_connector_options``: tuple of options passed to
  taktuk_connector.

- ``nc``: the netcat command to use

- ``chainput_nc_client_timeout``: timeout for client connection to
  next hop

- ``chainput_nc_server_timeout``: timeout for server to wait for
  incoming connection from previous hop

- ``chainput_host_retry``: number of times each hop in the transfer
  chain retries to connect to next hop

- ``chainput_chain_retry``: number of times each hop in the transfer
  chain tries a new next hop if all tries to current next hop fail. If
  given a float (between 0.0 and 1.0), this is expressed as a ratio of
  the total number of hosts in the chain.

- ``chainput_try_delay``: delay in seconds between TCP client to
  server connection attempts.

- ``forwarding_timeout``: max time to wait for a forwarding
  port to be available

- ``pty``: boolean. Wether to allocate or not a pty for
  ssh/scp.

- ``host_rewrite_func``: function called to rewrite hosts
  addresses. Takes a host address, returns a host address.
"""

def make_connection_params(connection_params = None, default_params = None):
    return_params = make_default_connection_params()
    if default_params == None:
        default_params = default_connection_params
    return_params.update(default_params)
    if connection_params: return_params.update(connection_params)
    return return_params

def load_configuration(filename, dicts_confs):
    """Update dicts with those found in file.

    :param filename: file to load dicts from

    :param dicts_confs: an iterable of couples (dict, string)

    Used to read configuration dicts. For each couple (dict, string),
    if a dict named string is defined in <file>, update
    dict with the content of this dict. Does nothing if unable to open
    <file>.
    """
    if os.path.isfile(filename):
        jailed_globals = {}
        try:
            with open(filename) as f:
                code = f.read()
                ccode = compile(code, filename, 'exec')
                exec(ccode, jailed_globals)
        except Exception as exc: #IGNORE:W0703
            print("ERROR while reading config file %s:" % (filename,))
            print(exc)
        for (dictio, conf) in dicts_confs:
            if conf in jailed_globals:
                dictio.update(jailed_globals[conf])

def get_user_config_filename():
    _user_conf_file = None
    if 'HOME' in os.environ:
        _user_conf_file = os.environ['HOME'] + '/.execo.conf.py'
    return _user_conf_file

__cf = get_user_config_filename()
if __cf:
    load_configuration(
        __cf,
        ((configuration, 'configuration'),
         (default_connection_params, 'default_connection_params')))



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

"""Execo extension and tools for Grid5000."""

from .config import g5k_configuration, \
    default_frontend_connection_params, default_oarsh_oarcp_params

from .oar import OarSubmission, oarsub, oardel, get_current_oar_jobs, \
    get_oar_job_info, wait_oar_job_start, get_oar_job_nodes, \
    get_oar_job_subnets, get_oar_job_kavlan, oarsubgrid

from .oargrid import oargridsub, oargriddel, \
    get_current_oargrid_jobs, get_oargrid_job_info, \
    get_oargrid_job_oar_jobs, wait_oargrid_job_start, \
    get_oargrid_job_nodes, get_oargrid_job_key

from .kadeploy import Deployment, Kadeployer, deploy

from .utils import get_kavlan_host_name, G5kAutoPortForwarder

from .planning import get_planning, compute_slots, find_first_slot, \
    find_max_slot, find_free_slot, get_jobs_specs, distribute_hosts, \
    g5k_charter_time

from .api_utils import get_g5k_sites, get_site_clusters, \
    get_cluster_hosts, get_g5k_clusters, get_g5k_hosts, \
    get_cluster_site, APIConnection, APIException, get_host_site, \
    get_host_cluster, group_hosts, get_resource_attributes, \
    get_site_network_equipments, get_host_attributes, get_cluster_attributes, \
    get_site_attributes, get_network_equipment_attributes, canonical_host_name,\
    get_network_equipment_site, get_host_shortname, get_host_longname, \
    get_cluster_network_equipments, get_site_hosts, get_host_network_equipments, \
    get_hosts_metric

from .charter import g5k_charter_time, get_next_charter_period

try:
    from .charter import g5k_charter_remaining
except:
    pass # no psycopg2

try:
    from .topology import g5k_graph, treemap
except:
    pass # no networkx



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .log import style
from .time_utils import format_date, format_duration
import sys
if sys.version_info >= (3,):
    _BIGNUM = sys.maxsize
else:
    _BIGNUM = sys.maxint

#def sort_reports(reports):
#    reports.sort(key = lambda report: report.stats().get('start_date') or _BIGNUM)

class Report(object):

    """Summary of one or more `execo.action.Action`.

    A Report gathers the results of actions or (recursively) of other
    reports.
    """

    def __init__(self, stats_objects = None):
        """
        :param stats_objects:
        """
        self.__name = "%s" % (self.__class__.__name__,)
        self._stats = Report.empty_stats()
        self._stats['name'] = self.__name
        if stats_objects:
            self.add(stats_objects)

    @property
    def name(self):
        """Name given to this report. If None, a default
        name will be given."""
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value
        self._stats['name'] = self.__name

    def add(self, stats_objects):
        """Add some sub-`execo.report.Report` or `execo.action.Action` to this report.

        :param stats_objects:
        """
        self._stats['sub_stats'].extend([o.stats() for o in stats_objects])

    @staticmethod
    def empty_stats():
        """Return a stats initialized to zero."""
        return {
            'name': None,
            'start_date': None,
            'end_date': None,
            'num_processes': 0,
            'num_started': 0,
            'num_ended': 0,
            'num_errors': 0,
            'num_timeouts': 0,
            'num_forced_kills': 0,
            'num_non_zero_exit_codes': 0,
            'num_expect_fail': 0,
            'num_write_error': 0,
            'num_ok': 0,
            'num_finished_ok': 0,
            'sub_stats': [],
            }

    @staticmethod
    def aggregate_stats(stats):
        aggstats = stats.copy()
        no_end_date = False
        for substats in stats['sub_stats']:
            for k in substats:
                if k == 'start_date':
                    if (substats[k] != None
                        and (aggstats[k] == None or substats[k] < aggstats[k])):
                        aggstats[k] = substats[k]
                elif k == 'end_date':
                    if substats[k] == None:
                        no_end_date = True
                    elif aggstats[k] == None or substats[k] > aggstats[k]:
                        aggstats[k] = substats[k]
                elif k in [
                    'num_processes',
                    'num_started',
                    'num_ended',
                    'num_errors',
                    'num_timeouts',
                    'num_forced_kills',
                    'num_non_zero_exit_codes',
                    'num_expect_fail',
                    'num_write_error',
                    'num_ok',
                    'num_finished_ok'
                    ]:
                    aggstats[k] += substats[k]
        if no_end_date:
            aggstats['end_date'] = None
        return aggstats

    def stats(self):
        """Return a dict summarizing the statistics of all `execo.action.Action` and sub-`execo.report.Report` registered to this report.

        This stats dict contains the following metrics:

        - ``start_date``: earliest start date (unix timestamp) of all
          `Action` or None if none have started yet.

        - ``end_date``: latest end date (unix timestamp) of all
          `Action` or None if not available (not all started, not all
          ended).

        - ``num_processes``: number of processes in all `Action`.

        - ``num_started``: number of processes that have started.

        - ``num_ended``: number of processes that have ended.

        - ``num_errors``: number of processes that went in error
          when started.

        - ``num_timeouts``: number of processes that had to be killed
          (SIGTERM) after reaching their timeout.

        - ``num_forced_kills``: number of processes that had to be
          forcibly killed (SIGKILL) after not responding for some
          time.

        - ``num_non_zero_exit_codes``: number of processes that ran
          correctly but whose return code was != 0.

        - ``num_expect_fail``: number of processes on which there was
          an expect failure, i.e. the expect search did not find a
          match before reaching the expect timeout or the stream
          eof/error.

        - ``num_write_error``: number of processes on which there was
          a write error.

        - ``num_ok``: number of processes which:

          - did not started

          - started and not yet ended

          - started and ended and did not went in error (or where
            launched with flag ignore_error) , did not timeout (or
            where launched with flag ignore_timeout), and had an exit
            code == 0 (or where launched with flag ignore_exit_code).

        - ``num_finished_ok``: number of processes which started,
          ended, and are ok.
        """
        return Report.aggregate_stats(self._stats)

    def __repr__(self):
        return "<Report(<%i entries>, name=%r)>" % (len(self._stats['sub_stats']), self._stats['name'])

    def __str__(self):
        stats = self.stats()
        return "<Report(<%i entries>, name=%r, start_date=%r, end_date=%r, num_processes=%r, num_started=%r, num_ended=%r, num_timeouts=%r, num_errors=%r, num_forced_kills=%r, num_expect_fail=%r, num_write_error=%r, num_non_zero_exit_codes=%r, num_ok=%r, num_finished_ok=%r)>" % (
            len(stats['sub_stats']),
            stats['name'],
            format_date(stats['start_date']),
            format_date(stats['end_date']),
            stats['num_processes'],
            stats['num_started'],
            stats['num_ended'],
            stats['num_timeouts'],
            stats['num_errors'],
            stats['num_forced_kills'],
            stats['num_expect_fail'],
            stats['num_write_error'],
            stats['num_non_zero_exit_codes'],
            stats['num_ok'],
            stats['num_finished_ok'])

    def to_string(self, wide = False, brief = False):
        """Returns a formatted string with a human-readable stats of all `Action` results.

        :param wide: if False (default), report format is designed for
          80 columns display. If True, output a (175 characters) wide
          report.

        :param brief: when True, only the Total stats is output, not
          each `Action` or `Report` stats. Default is False.
        """
        output = ""
        if wide:
            output += "Name                            start               end                 length         startd ended  error  timout fkilld expfld wrterr ret!=0 ok     total  \n"
            output += "-------------------------------------------------------------------------------------------------------------------------------------------------------------\n"
        else:
            output += "Name                            start               end                 total  \n"
            output += " length         startd ended  error  timout fkilld expfld wrterr ret!=0 ok     \n"
            output += "-------------------------------------------------------------------------------\n"
        def format_line(stats, indent):
            result = ""
            indented_name = " " * indent + "%s" % (stats.get('name'),)
            length = ""
            if stats.get('start_date') and stats.get('end_date'):
                length = format_duration(stats['end_date'] - stats['start_date'])
            else:
                length = ""
            if wide:
                tmpline = "%-31.31s %-19.19s %-19.19s %-15.15s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s\n" % (
                    indented_name,
                    format_date(stats['start_date']),
                    format_date(stats['end_date']),
                    length,
                    stats['num_started'],
                    stats['num_ended'],
                    stats['num_errors'],
                    stats['num_timeouts'],
                    stats['num_forced_kills'],
                    stats['num_expect_fail'],
                    stats['num_write_error'],
                    stats['num_non_zero_exit_codes'],
                    stats['num_ok'],
                    stats['num_processes'])
            else:
                tmpline = "%-31.31s %-19.19s %-19.19s %-7.7s\n" % (
                    indented_name,
                    format_date(stats['start_date']),
                    format_date(stats['end_date']),
                    stats['num_processes'],)
                tmpline += " %-15.15s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s%-7.7s\n" % (
                    length,
                    stats['num_started'],
                    stats['num_ended'],
                    stats['num_errors'],
                    stats['num_timeouts'],
                    stats['num_forced_kills'],
                    stats['num_expect_fail'],
                    stats['num_write_error'],
                    stats['num_non_zero_exit_codes'],
                    stats['num_ok'],)
            if stats['num_ok'] < stats['num_processes']:
                if stats['num_ok'] == stats['num_ended'] and (stats['num_expect_fail'] == 0 and stats['num_write_error'] == 0):
                    tmpline = style.report_warn(tmpline)
                else:
                    tmpline = style.report_error(tmpline)
            result += tmpline
            return result

        def recurse_stats(stats, indent):
            result = ""
            result += format_line(stats, indent)
            for sub_stats in stats['sub_stats']:
                result += recurse_stats(sub_stats, indent+2)
            return result

        stats = self.stats()
        if not brief and len(stats['sub_stats']) > 0:
            for sub_stats in sorted(stats['sub_stats'],
                                    key = lambda stats: stats.get('start_date') or _BIGNUM):
                output += recurse_stats(sub_stats, 0)
            if wide:
                output += "--------------------------------------------------------------------------------------------------------------------------------------------------------------\n"
            else:
                output += "-------------------------------------------------------------------------------\n"

        output += format_line(stats, 0)
        return output



from execo import *
from execo_g5k import *
import tempfile, shutil

logger.info("compute resources to reserve")
slots = compute_slots(get_planning(), 60*15)
wanted = { "grid5000": 0 }
start_date, end_date, resources = find_first_slot(slots, wanted)
actual_resources = distribute_hosts(resources, wanted, ratio = 0.9)
job_specs = get_jobs_specs(actual_resources)
logger.info("try to reserve " + str(actual_resources))
jobid, sshkey = oargridsub(job_specs, start_date,
                           walltime = end_date - start_date)
if jobid:
    try:
        logger.info("generate random data")
        Process("dd if=/dev/urandom of=randomdata bs=1M count=50").run()
        logger.info("wait job start")
        wait_oargrid_job_start(jobid)
        logger.info("get job nodes")
        nodes = get_oargrid_job_nodes(jobid)
        logger.info("got %i nodes" % (len(nodes),))
        conn_parms = default_oarsh_oarcp_params.copy()
        tmpsshkey = tempfile.mktemp()
        shutil.copy2(sshkey, tmpsshkey)
        conn_parms['taktuk_options'] = ( "-s", "-S", "%s:%s" % (tmpsshkey, tmpsshkey))
        conn_parms['keyfile'] = tmpsshkey
        broadcast1 = ChainPut(nodes, ["randomdata"], "/tmp/",
                              connection_params = conn_parms)
        broadcast2 = Put(nodes, ["randomdata"], "/tmp/",
                         connection_params = conn_parms)
        logger.info("run chainput")
        broadcast1.run()
        logger.info("run parallel scp")
        broadcast2.run()
        logger.info("summary:\n" + Report([broadcast1, broadcast2]).to_string())
    finally:
        logger.info("deleting job")
        oargriddel([jobid])



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

"""Handles launch and control of several OS level processes concurrently.

Handles remote executions and file copies with ssh or similar tools.
"""

from .log import logger
from .config import configuration, default_connection_params, \
  SSH, TAKTUK, SCP, CHAINPUT
from .time_utils import sleep, Timer, format_date, format_duration, \
  get_seconds, get_unixts
from .host import Host
from .process import Process, SshProcess, get_process, \
     ProcessLifecycleHandler, ProcessOutputHandler, \
     PortForwarder, Serial, SerialSsh, STDOUT, STDERR, \
     ExpectOutputHandler
from .action import Action, wait_any_actions, wait_all_actions, \
  Remote, Put, Get, TaktukRemote, TaktukPut, TaktukGet, Local, \
  ParallelActions, SequentialActions, default_action_factory, \
  get_remote, get_fileput, get_fileget, \
  ActionLifecycleHandler, ChainPut, filter_bad_hosts, \
  RemoteSerial
from .report import Report
from .exception import ProcessesFailed, ActionsFailed
try:
    from ._version import __version__
except:
    pass



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .log import logger
import datetime, calendar, time, re

def timedelta_to_seconds(td):
    """Convert a ``datetime.timedelta`` to a number of seconds (float)."""
    return td.days * 86400 + td.seconds + td.microseconds / 1e6

def datetime_to_unixts(dt):
    """Convert a ``datetime.datetime`` to a utc unix timestamp (float).

    If the datetime given has an explicit timezone, it is used,
    otherwise the timestamp is assumed to be given in the locale
    timezone and is converted to utc.
    """
    if dt.tzinfo:
        return calendar.timegm(dt.utctimetuple()) + dt.microsecond / 1e6
    else:
        return time.mktime(dt.timetuple()) + dt.microsecond / 1e6

def unixts_to_datetime(ts, tz = None):
    """Convert a utc unixts (int or float) to ``datetime.datetime``.

    If no timezone is given, the returned datetime is a naive one,
    with the timestamp set in local time, without a timezone. Or the
    returned datetime is set in the given timezone if one is given.
    """
    return datetime.datetime.fromtimestamp(ts, tz)

def numeric_date_to_unixts(d):
    """Convert a numeric to unix timestamp (float)."""
    return float(d)

def numeric_duration_to_seconds(d):
    """Convert a numeric to duration in number of seconds (float)."""
    return float(d)

_num_str_date_re = re.compile("^((\d*\.)?\d+)$")
_str_date_re = re.compile("^\s*((((\d\d)?(\d\d)-)?((\d?\d)-))?(\d?\d)[Tt ])?(\d?\d):(\d?\d):(\d?\d)(\.\d+)?([Zz]|([+-])(\d?\d):(\d?\d))?\s*$")
def str_date_to_unixts(d):
    """Convert a date string to a unix timestamp (float).

    date string format can be:

    - a numeric. In this case, convert with
      `execo.time_utils.numeric_date_to_unixts`

    - a string in rfc-3339 format 'YYYY-MM-DD[Tt ]HH:MM:SS[.subsec][Zz|(+|-)HH:MM]'.
      If a timezone is given, it is used, otherwise the timestamp is
      assumed to be given in the locale timezone and is converted to
      utc.
    """
    numeric_date = _num_str_date_re.match(d)
    if numeric_date:
        return numeric_date_to_unixts(float(numeric_date.group(1)))
    str_date = _str_date_re.match(d)
    if str_date:
        now = time.localtime()
        if str_date.group(5):
            year = int(str_date.group(5))
            if str_date.group(4):
                year += 100 * int(str_date.group(4))
            else:
                if year >= 69: year += 1900
                else: year += 2000
        else:
            year = now.tm_year
        if str_date.group(7):
            month = int(str_date.group(7))
        else:
            month = now.tm_mon
        if str_date.group(8):
            day = int(str_date.group(8))
        else:
            day = now.tm_mday
        h = int(str_date.group(9))
        m = int(str_date.group(10))
        s = int(str_date.group(11))
        norm_date = "%i-%i-%i %i:%i:%i" % (year, month, day, h, m, s)
        if str_date.group(13):
            ts = calendar.timegm(time.strptime(norm_date, "%Y-%m-%d %H:%M:%S"))
        else:
            ts = time.mktime(time.strptime(norm_date, "%Y-%m-%d %H:%M:%S"))
        if str_date.group(12):
            ts += float(str_date.group(12))
        if str_date.group(13) and str_date.group(13) not in ["Z", "z"]:
            offset = int(str_date.group(15)) * 3600 + int(str_date.group(16)) * 60
            if str_date.group(14) == "+": ts -= offset
            else: ts += offset
        return ts
    raise ValueError("unsupported date format %s" % (d,))

_num_str_duration_re = re.compile("^((\d*\.)?\d+)$")
_str_duration_re = re.compile("^(\d+):(\d?\d):(\d?\d)(\.\d+)?$")
def str_duration_to_seconds(duration):
    """Convert a duration string to a number of seconds (float).

    duration string format can be:

    - a numeric. In this case, convert with
      `execo.time_utils.numeric_duration_to_seconds`

    - a string in format 'HH:MM:SS[.MS]'
    """
    numeric_duration = _num_str_duration_re.match(duration)
    if numeric_duration:
        return numeric_duration_to_seconds(float(numeric_duration.group(1)))
    str_duration = _str_duration_re.match(duration)
    if str_duration:
        duration = (int(str_duration.group(1)) * 3600
                    + int(str_duration.group(2)) * 60
                    + int(str_duration.group(3)))
        if str_duration.group(4):
            duration += float(str_duration.group(4))
        return duration
    raise ValueError("unsupported duration format %s" % (duration,))

def get_seconds(duration):
    """Convert a duration to a number of seconds (float).

    :param duration: a duration in one of the supported types. if
      duration == None, returns None. Supported types

      - ``datetime.timedelta``: see `execo.time_utils.timedelta_to_seconds`

      - string: see `execo.time_utils.str_duration_to_seconds`

      - numeric type: see
        `execo.time_utils.numeric_duration_to_seconds`
    """
    if duration == None:
        return None
    elif isinstance(duration, datetime.timedelta):
        return timedelta_to_seconds(duration)
    elif isinstance(duration, str):
        return str_duration_to_seconds(duration)
    return numeric_duration_to_seconds(duration)

def get_unixts(d):
    """Convert a date to a unix timestamp (float).

    :param d: a date in one of the supported types. if date == None,
      returns None. Supported types

      - ``datetime.datetime``: see `execo.time_utils.datetime_to_unixts`

      - string: see `execo.time_utils.str_date_to_unixts`

      - numeric type: see `execo.time_utils.numeric_date_to_unixts`
    """
    if d == None:
        return None
    if isinstance(d, datetime.datetime):
        return datetime_to_unixts(d)
    elif isinstance(d, str):
        return str_date_to_unixts(d)
    return numeric_date_to_unixts(d)

def _get_milliseconds_suffix(secs):
    """Return a formatted millisecond suffix, either empty if ms = 0, or dot with 3 digits otherwise.

    :param secs: a unix timestamp (integer or float)
    """
    ms_suffix = ""
    msecs = int (round(secs - int(secs), 3) * 1000)
    if msecs != 0:
        ms_suffix = ".%03i" % msecs
    return ms_suffix

def _zone3339(timetuple):
    """Return a string with the locale timezone in rfc-3339 format."""
    dst = timetuple[8]
    offs = (time.timezone, time.timezone, time.altzone)[1 + dst]
    return '%+.2d:%.2d' % (offs // -3600, abs(offs // 60) % 60)

def format_unixts(secs, showms = False):
    """Return a string with the formatted date (year, month, day, hour, min, sec, ms) in locale timezone and in rfc-3339 format for pretty printing.

    :param secs: a unix timestamp (integer or float) (or None).

    :param showms: whether to show ms or not. Default False.
    """
    if secs == None:
        return None
    t = time.localtime(secs)
    formatted_time = time.strftime("%Y-%m-%d %H:%M:%S", t)
    if showms:
        formatted_time += _get_milliseconds_suffix(secs)
    formatted_time += _zone3339(t)
    return formatted_time

def format_seconds(secs, showms = False):
    """Return a string with a formatted duration (days, hours, mins, secs, ms) for pretty printing.

    :param secs: a duration in seconds (integer or float) (or None).

    :param showms: whether to show ms or not. Default False.
    """
    if secs == None:
        return None
    formatted_duration = ""
    if secs < 0:
        secs = -secs
        formatted_duration += "-"
    s = secs
    d = (s - (s % 86400)) // 86400
    s -= d * 86400
    h = (s - (s % 3600)) // 3600
    s -= h * 3600
    m = (s - (s % 60)) // 60
    s -= m * 60
    if secs >= 86400: formatted_duration += "%id" % d
    if secs >= 3600: formatted_duration += "%ih" % h
    if secs >= 60: formatted_duration += "%im" % m
    if showms:
        formatted_duration += "%i%ss" % (s, _get_milliseconds_suffix(s))
    else:
        formatted_duration += "%is" % (s,)
    return formatted_duration

def format_date(d, showms = False):
    """Return a string with the formatted date (year, month, day, hour, min, sec, ms) in locale timezone and in rfc-3339 format for pretty printing.

    :param d: a date in one of the formats handled (or None) (see
      `execo.time_utils.get_unixts`).

    :param showms: whether to show ms or not. Default False.
    """
    return format_unixts(get_unixts(d), showms)

def format_duration(duration, showms = False):
    """Return a string with a formatted duration (days, hours, mins, secs, ms) for pretty printing.

    :param duration: a duration in one of the formats handled (or
      None) (see `execo.time_utils.get_seconds`).

    :param showms: whether to show ms or not. Default False.
    """
    return format_seconds(get_seconds(duration), showms)

def _safe_sleep(secs):
    """Safe sleeping: restarted if interrupted by signals.

    :param secs: time to sleep in seconds (int or float)
    """
    end = time.time() + secs
    sleep_time = secs
    while sleep_time > 0:
        time.sleep(sleep_time)
        sleep_time = end - time.time()

def sleep(delay = None, until = None):
    """Sleep until a given delay has elapsed or until a given date.

    If both present, will sleep at least for the delay and at least
    until the date.

    :param delay: the delay to sleep in one of the formats handled (or
      None) (see `execo.time_utils.get_seconds`).

    :param until: the date until which to sleep in one of the formats
      handled (or None) (see `execo.time_utils.get_unixts`).
    """
    delay = get_seconds(delay)
    until = get_unixts(until)
    sleeptime = 0
    if delay != None:
        sleeptime = delay
    if until != None:
        dt = until - time.time()
        if (sleeptime > 0 and dt > sleeptime) or (sleeptime <= 0 and dt > 0):
            sleeptime = dt
    if sleeptime > 0:
        logger.debug("sleeping %s", format_seconds(sleeptime))
        _safe_sleep(sleeptime)
        logger.debug("end sleeping")
        return sleeptime

class Timer(object):

    """Keep track of elapsed time."""

    def __init__(self, timeout = None):
        """Create and start the Timer."""
        self._start = time.time()
        self._timeout = get_seconds(timeout)

    def wait_elapsed(self, elapsed):
        """Sleep until the given amount of time has elapsed since the Timer's start.

        :param elapsed: the delay to sleep in one of the formats
          handled (or None) (see `execo.time_utils.get_seconds`).
        """
        elapsed = get_seconds(elapsed)
        really_elapsed = time.time() - self._start
        if really_elapsed < elapsed:
            sleep(elapsed - really_elapsed)
        return self

    def start_date(self):
        """Return this Timer's instance start time."""
        return self._start

    def elapsed(self):
        """Return this Timer's instance elapsed time since start."""
        return time.time() - self._start

    def remaining(self):
        """Returns the remaining time before the timeout."""
        if self._timeout != None:
            return self._start + self._timeout - time.time()
        else:
            return None



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .log import logger
import os, sys, time, inspect, pipes
from .utils import redirect_outputs, copy_outputs
from argparse import ArgumentParser

_engineargs = sys.argv[1:]

def run_meth_on_engine_ancestors(instance, method_name):
    engine_ancestors = [ cls for cls in inspect.getmro(instance.__class__) if issubclass(cls, Engine) ]
    for cls in engine_ancestors:
        meth = cls.__dict__.get(method_name)
        if meth is not None:
            meth(instance)

class Engine(object):

    """Basic class for execo Engine.

    Subclass it to develop your own engines, possibly reusable.

    This class offers basic facilities:

    - central handling of options and arguments

    - automatic experiment directory creation

    - various ways to handle stdout / stderr

    - support for continuing a previously stopped experiment

    - log level selection

    A subclass of Engine can access the following member variables
    which are automatically defined and initialized at the right time
    by the base class `execo_engine.engine.Engine`:

    - `execo_engine.engine.Engine.engine_dir`

    - `execo_engine.engine.Engine.result_dir`

    - `execo_engine.engine.Engine.args_parser`

    - `execo_engine.engine.Engine.args`

    - `execo_engine.engine.Engine.run_name`

    - `execo_engine.engine.Engine.result_dir`

    A subclass of Engine can override the following methods:

    - `execo_engine.engine.Engine.init`

    - `execo_engine.engine.Engine.run`

    - `execo_engine.engine.Engine.setup_run_name`

    - `execo_engine.engine.Engine.setup_result_dir`

    A typical, non-reusable engine would start by adding options /
    arguments to `execo_engine.engine.Engine.args_parser` in `__init__()`,
    then override `execo_engine.engine.Engine.init` to perform further
    initialization if needed.  It would then implement all the experiment
    code by overriding `execo_engine.engine.Engine.run`.  This ensures
    that all initialization steps are performed by the engine before the
    experiment runs: results directory is initialized and created
    (possibly reusing a previous results directory, to restart from a
    previously stopped experiment), log level is set, stdout / stderr are
    redirected as needed, and options and arguments are in
    `execo_engine.engine.Engine.args`.

    Example engine with custom command-line arguments::

     class MyEngine(execo_engine.Engine):
         def __init__(self):
             super(MyEngine, self).__init__()
             self.args_parser.add_argument('--myoption', default='foo',
                                           help='An option to control how the experiment is done')
         def run(self):
             if self.args.myoption == "foo":
                 ...

    A typical usage of a `execo_engine.utils.ParamSweeper` in an
    engine would be to initialize an instance at the beginning of
    `execo_engine.engine.Engine.run`, using a persistent file in the
    results directory. Example code::

     def run(self):
         sweeps = sweep({<parameters and their values>})
         sweeper = ParamSweeper(sweeps, os.path.join(self.result_dir, "sweeps"))
         [...]

    """

    def _create_result_dir(self):
        """Ensure the engine's result dir exists. Create it if needed."""
        try:
            os.makedirs(self.result_dir)
        except os.error:
            pass

    def __init__(self):
        self.engine_dir = None
        """Full path of the engine directory, if available. May not be
        available if engine code is run interactively.
        """
        mymodule = sys.modules[self.__module__]
        if hasattr(mymodule, "__file__"):
            self.engine_dir = os.path.abspath(os.path.dirname(os.path.realpath(mymodule.__file__)))
        self.args_parser = ArgumentParser(usage = "<program> [options] <arguments>",
                                          description = "engine: " + self.__class__.__name__)
        """Subclasses of `execo_engine.engine.Engine` can register options and args to this options parser in `__init__()`."""
        self.args_parser.add_argument(
            "-l", dest = "log_level", default = None,
            help = "log level (int or string). Default = inherit execo logger level")
        self.args_parser.add_argument(
            "-L", dest = "output_mode", action="store_const", const = "copy", default = False,
            help = "copy stdout / stderr to log files in the experiment result directory. Default = %(default)s")
        self.args_parser.add_argument(
            "-R", dest = "output_mode", action="store_const", const = "redirect", default = False,
            help = "redirect stdout / stderr to log files in the experiment result directory. Default = %(default)s")
        self.args_parser.add_argument(
            "-M", dest = "merge_outputs", action="store_true", default = False,
            help = "when copying or redirecting outputs, merge stdout / stderr in a single file. Default = %(default)s")
        self.args_parser.add_argument(
            "-c", dest = "use_dir", default = None, metavar = "DIR",
            help = "use experiment directory DIR")
        self.args = None
        """Arguments and options given on the command line. Available after
        the command line has been parsed, in
        `execo_engine.engine.Engine.run` (not in
        `execo_engine.engine.Engine.init`)

        """
        self.run_name = None
        """Name of the current experiment. If you want to modify it,
        override `execo_engine.engine.Engine.setup_run_name`
        """
        self.result_dir = None
        """full path to the current engine's execution results
        directory, where results should be written, where stdout /
        stderr are output, where `execo_engine.utils.ParamSweeper`
        persistence files should be written, and where more generally
        any file pertaining to a particular execution of the
        experiment should be located.
        """

    def start(self, engineargs = _engineargs):
        """Start the engine.

        Properly initialize the experiment Engine instance, then call
        the init() method of all subclasses, then pass the control to
        the overridden run() method of the requested experiment
        Engine.
        """
        self.args = self.args_parser.parse_args(args = engineargs)
        # _engineargs hack: to make running an engine from ipython
        # more convenient: if ipython is run from command line as
        #
        # $ ipython -i <enginefile.py> -- [engine options]
        #
        # iptyhon will hack sys.argv when loading enginefile, hiding
        # its own args, but then in the interactive shell, we get back
        # the real sys.argv, including ipython args, so if we
        # instanciate or start the engine interactively, we need to
        # parse the right args from sys.argv which is thus saved at
        # engine load time.
        if self.args.log_level != None:
            try:
                log_level = int(self.args.log_level)
            except:
                log_level = self.args.log_level
            logger.setLevel(log_level)
        if sys.version_info < (2, 7):
            if len(self.args) < self.args_parser.num_arguments():
                self.args_parser.print_help(sys.stderr)
                exit(1)
        self.setup_run_name()
        if self.args.use_dir:
            self.result_dir = self.args.use_dir
        else:
            self.setup_result_dir()
        self._create_result_dir()
        if self.args.output_mode:
            if self.args.merge_outputs:
                stdout_fname = self.result_dir + "/stdout+stderr"
                stderr_fname = self.result_dir + "/stdout+stderr"
            else:
                stdout_fname = self.result_dir + "/stdout"
                stderr_fname = self.result_dir + "/stderr"
            if self.args.output_mode == "copy":
                copy_outputs(stdout_fname, stderr_fname)
                logger.info("dup stdout / stderr to %s and %s", stdout_fname, stderr_fname)
            elif self.args.output_mode == "redirect":
                redirect_outputs(stdout_fname, stderr_fname)
                logger.info("redirect stdout / stderr to %s and %s", stdout_fname, stderr_fname)
        logger.info("command line arguments: %s" % (sys.argv,))
        logger.info("command line: " + " ".join([pipes.quote(arg) for arg in sys.argv]))
        logger.info("run in directory %s", self.result_dir)
        run_meth_on_engine_ancestors(self, "init")
        run_meth_on_engine_ancestors(self, "run")

    # ------------------------------------------------------------------
    #
    # below: methods that inherited real experiment engines can
    # override
    #
    # ------------------------------------------------------------------

    def setup_run_name(self):
        """Set the experiment run name.

        Default implementation: concatenation of class name and
        date. Override this method to change the name of the
        experiment. This method is called before
        `execo_engine.engine.Engine.run` and
        `execo_engine.engine.Engine.init`
        """
        self.run_name = self.__class__.__name__ + "_" + time.strftime("%Y%m%d_%H%M%S_%z")

    def setup_result_dir(self):
        """Set the experiment run result directory name.

        Default implementation: subdirectory with the experiment run
        name in the current directory. Override this method to change
        the name of the result directory. This method is called before
        `execo_engine.engine.Engine.run` and
        `execo_engine.engine.Engine.init`. Note that if option ``-c``
        is given to the engine, the name given on command line will
        take precedence, and this method won't be called.
        """
        self.result_dir = os.path.abspath(self.run_name)

    def init(self):
        """Experiment init method

        Override this method with the experiment init code. Default
        implementation does nothing.

        The base class `execo_engine.engine.Engine` takes care that
        all `execo_engine.engine.Engine.init` methods of its subclass
        hierarchy are called, in the order ancestor method before
        subclass method. This order is chosen so that generic engines
        inheriting from `execo_engine.engine.Engine` can easily
        implement common functionnalities. For example a generic
        engine can declare its own options and arguments in ``init``,
        which will be executed before a particular experiment subclass
        ``init`` method.
        """
        pass

    def run(self):
        """Experiment run method

        Override this method with the experiment code. Default
        implementation does nothing.

        The base class `execo_engine.engine.Engine` takes care that
        all `execo_engine.engine.Engine.run` methods of its subclass
        hierarchy are called, in the order ancestor method before
        subclass method. This order is chosen so that generic engines
        inheriting from `execo_engine.engine.Engine` can easily
        implement common functionnalities. For example a generic
        engine can prepare an experiment environment in its ``run``
        method, which will be executed before a particular experiment
        subclass ``run`` method.
        """
        pass



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>


from .log import logger
from .engine import Engine
from .utils import slugify, redirect_outputs, copy_outputs
from .sweep import HashableDict, sweep, ParamSweeper, geom, igeom, sweep_stats



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from execo.config import make_connection_params, configuration, SSH, TAKTUK, SCP, CHAINPUT
from execo.host import Host
from execo.process import get_process, STDOUT, STDERR, ExpectOutputHandler
from .host import get_hosts_list, get_unique_hosts_list
from .log import style, logger
from .process import ProcessLifecycleHandler, SshProcess, ProcessOutputHandler, \
    TaktukProcess, Process, SerialSsh
from .report import Report
from .ssh_utils import get_rewritten_host_address, get_scp_command, \
    get_taktuk_connector_command, get_ssh_command
from .utils import name_from_cmdline, non_retrying_intr_cond_wait, intr_event_wait, get_port, \
    singleton_to_collection
from traceback import format_exc
from .substitutions import get_caller_context, remote_substitute
from .time_utils import get_seconds, format_date, Timer
import threading, time, pipes, tempfile, os, shutil, stat

class ActionLifecycleHandler(object):

    """Abstract handler for `execo.action.Action` lifecycle."""

    def start(self, action):
        """Handle `execo.action.Action`'s start.

        :param action: The Action which starts.
        """
        pass

    def end(self, action):
        """Handle `execo.action.Action`'s end.

        :param action: The Action which ends.
        """
        pass

    def reset(self, action):
        """Handle `execo.action.Action`'s reset.

        :param action: the Action which is reset.
        """
        pass

class Action(object):

    """Abstract base class. A set of parallel processes.

    An `execo.action.Action` can be started
    (`execo.action.Action.start`), killed
    (`execo.action.Action.kill`). One can wait
    (`execo.action.Action.wait`) for an Action, it means waiting for
    all processes in the Action to finish. An Action can be run
    (`execo.action.Action.run`), it means start it then wait for it to
    complete.

    An Action and its subclasses can act as a context manager object,
    allowing to write code such as::

     with Remote(...).start() as r:
       [...do something...]

    When exiting the contex manager scope, the remote is automatically
    killed.
    """

    _wait_multiple_actions_condition = threading.Condition()

    def __init__(self, lifecycle_handlers = None, name = None, default_expect_timeout = None):
        """:param lifecycle_handlers: List of instances of
          `execo.action.ActionLifecycleHandler` for being notified of
          action lifecycle events.

        :param name: User-friendly name. A default is generated and
          can be changed.

        :param default_expect_timeout: The default timeout for expect
          invocations when no explicit timeout is given. Defaults to
          None, meaning no timeout.
        """
        self.started = False
        """whether this Action was started (boolean)"""
        self.ended = False
        """whether this Action has ended (boolean)"""
        # self.timeout = timeout
        # """Timeout in seconds, or None"""
        # self.factory = factory
        # """Factory which created this action, or None"""
        self.processes = []
        """Iterable of all `execo.process.ProcessBase`"""
        if name != None:
            self.name = name
            """User-friendly name. A default is generated and can be changed."""
        else:
            self.name = "%s 0x%08.8x" % (self.__class__.__name__, id(self))
        self.default_expect_timeout = default_expect_timeout
        """The default timeout for expect invocations when no explicit timeout
        is given. Defaults to None, meaning no timeout."""
        if lifecycle_handlers != None:
            self.lifecycle_handlers = list(lifecycle_handlers)
            """List of instances of `execo.action.ActionLifecycleHandler` for
            being notified of action lifecycle events.
            """
        else:
            self.lifecycle_handlers = list()
        self._end_event = threading.Event()
        self._end_event.clear()
        self._thread_local_storage = threading.local()
        self._thread_local_storage.expect_handler = None

    def _common_reset(self):
        # all methods _common_reset() of this class hierarchy contain
        # the code common to the constructor and reset() to
        # reinitialize an object. _common_reset() of child classes
        # must explicitely call _common_reset() of their parent class.
        self.started = False
        self.ended = False
        self._end_event.clear()
        self._thread_local_storage.expect_handler = None
        self._init_processes()

    def _args(self):
        # to be implemented in all subclasses. Must return a list with
        # all arguments to the constructor, beginning by the
        # positionnal arguments, finishing by keyword arguments. This
        # list will be directly used in __repr__ methods.
        return Action._kwargs(self)

    def _kwargs(self):
        # to be implemented in all subclasses. Must return a list with
        # all keyword arguments to the constructor. This list will be
        # used to build the list returned by _args() of this class or
        # child classes.
        kwargs = []
        if self.default_expect_timeout != None: kwargs.append("default_expect_timeout=%r" % (self.default_expect_timeout,))
        return kwargs

    def _infos(self):
        # to be implemented in all subclasses. Must return a list with
        # all relevant infos other than those returned by _args(), for
        # use in __str__ methods.
        stats = self.stats()
        infos = [ "name=%s" % (self.name,),
                  "started=%r" % (self.started,),
                  "start_date=%r" % (format_date(stats['start_date']),),
                  "ended=%r" % (self.ended,),
                  "end_date=%r" % (format_date(stats['end_date']),),
                  "num_processes=%r" % (stats['num_processes'],),
                  "num_started=%r" % (stats['num_started'],),
                  "num_ended=%r" % (stats['num_ended'],),
                  "num_timeouts=%r" % (stats['num_timeouts'],),
                  "num_errors=%r" % (stats['num_errors'],),
                  "num_forced_kills=%r" % (stats['num_forced_kills'],),
                  "num_non_zero_exit_codes=%r" % (stats['num_non_zero_exit_codes'],),
                  "num_ok=%r" % (stats['num_ok'],),
                  "num_finished_ok=%r" % (stats['num_finished_ok'],),
                  "ok=%r" % (self.ok,) ]
        return infos

    def __repr__(self):
        # implemented once for all subclasses
        return "%s(%s)" % (self.__class__.__name__, ", ".join(self._args()))

    def __str__(self):
        # implemented once for all subclasses
        return "<" + style.object_repr(self.__class__.__name__) + "(%s)>" % (", ".join(self._args() + self._infos()),)

    def _notify_terminated(self):
        with Action._wait_multiple_actions_condition:
            logger.debug(style.emph("got termination notification for:") + " %s", self)
            for handler in list(self.lifecycle_handlers):
                try:
                    handler.end(self)
                except Exception as e:
                    logger.error("action lifecycle handler %s end raised exception for action %s:\n%s" % (
                            handler, self, format_exc()))
            self.ended = True
            self._end_event.set()
            Action._wait_multiple_actions_condition.notifyAll()

    def _init_processes(self):
        self.processes = []

    def start(self):
        """Start all processes.

        return self"""
        if self.started:
            raise ValueError("Actions may be started only once")
        self.started = True
        logger.debug(style.emph("start:") + " %s", self)
        for handler in list(self.lifecycle_handlers):
            try:
                handler.start(self)
            except Exception as e:
                logger.error("action lifecycle handler %s start raised exception for action %s:\n%s" % (
                        handler, self, format_exc()))
        return self

    def kill(self):
        """Kill all processes not yet ended.

        Returns immediately, without waiting for processes to be
        actually killed.

        return self"""
        logger.debug(style.emph("kill:") + " %s", self)
        return self

    def wait(self, timeout = None):
        """Wait for all processes to complete.

        return self"""
        logger.debug(style.emph("start waiting:") + " %s", self)
        intr_event_wait(self._end_event, get_seconds(timeout))
        logger.debug(style.emph("end waiting:") + " %s", self)
        return self

    def run(self, timeout = None):
        """Start all processes then wait for them to complete.

        return self"""
        logger.debug(style.emph("run:") + " %s", self)
        self.start()
        self.wait(timeout)
        return self

    def reset(self):
        """Reinitialize an Action so that it can later be restarted.

        If it is running, this method will first kill it then wait for
        its termination before reseting.
        """
        logger.debug(style.emph("reset:") + " %s", self)
        if self.started and not self.ended:
            self.kill()
            self.wait()
        for handler in list(self.lifecycle_handlers):
            try:
                handler.reset(self)
            except Exception as e:
                logger.error("action lifecycle handler %s reset raised exception for action %s:\n%s" % (
                        handler, self, format_exc()))
        self._common_reset()
        return self

    @property
    def ok(self):
        """Returns a boolean indicating if all processes are ok.

        refer to `execo.process.ProcessBase.ok` for detailed semantics
        of being ok for a process.
        """
        return all([process.ok for process in self.processes])

    @property
    def finished_ok(self):
        """Action has started, ended, and is not in error."""
        return self.started and self.ended and self.ok

    def stats(self):
        """Return a dict summarizing the statistics of all processes
        of this Action.

        see `execo.report.Report.stats`.
        """
        stats = Report.empty_stats()
        stats['name'] = self.name
        for process in self.processes:
            pstats = process.stats()
            if (stats['start_date'] == None
                or (pstats['start_date'] != None
                    and pstats['start_date'] < stats['start_date'])):
                stats['start_date'] = pstats['start_date']
            if (stats['end_date'] == None
                or (pstats['end_date'] != None
                    and pstats['end_date'] > stats['end_date'])):
                stats['end_date'] = pstats['end_date']
            stats['num_processes'] += pstats['num_processes']
            stats['num_started'] += pstats['num_started']
            stats['num_ended'] += pstats['num_ended']
            stats['num_errors'] += pstats['num_errors']
            stats['num_timeouts'] += pstats['num_timeouts']
            stats['num_forced_kills'] += pstats['num_forced_kills']
            stats['num_non_zero_exit_codes'] += pstats['num_non_zero_exit_codes']
            stats['num_ok'] += pstats['num_ok']
            stats['num_finished_ok'] += pstats['num_finished_ok']
        if stats['num_processes'] > stats['num_ended']:
            stats['end_date'] = None
        return stats

    def __enter__(self):
        """Context manager enter function: returns self"""
        return self

    def __exit__(self, t, v, traceback):
        """Context manager leave function: kills the action"""
        self.kill()
        return False

    def expect(self, regexes, timeout = False, stream = STDOUT, backtrack_size = 2000, start_from_current = False):
        """searches the process output stream(s) for some regex. It mimics/takes ideas from Don Libes expect, or python-pexpect, but in parallel on several processes.

        It is an easier-to-use frontend for
        `execo.process.ExpectOutputHandler`.

        It waits for a regex to match on all processes. Then it
        returns a list of tuples (process, regex index, match
        object). For processes / streams for which there was no match
        before reaching the timeout or the stream is eof or error, the
        tuple is (process, None, None). The returned list has the same
        process sort order than self.processes.

        It uses thread local storage such that concurrent expects in
        parallel threads do not interfere which each other.

        :param regexes: a regex or list of regexes. May be given as string
          or as compiled regexes.

        :param timeout: wait timeout after which it returns (None,
          None) if no match was found. If False (the default): use the
          default expect timeout. If None: no timeout.

        :param stream: stream to monitor for this process, STDOUT or
          STDERR.

        :param backtrack_size: Each time some data is received, this
          ouput handler needs to perform the regex search not only on
          the incoming data, but also on the previously received data,
          or at least on the last n bytes of the previously received
          data, because the regex may match on a boundary between what
          was received in a previous read and what is received in the
          incoming read. These n bytes are the backtrack_size. (for
          special cases: if backtrack_size == None, the regex search
          is always done on the whole received data, but beware, this
          is probably not what you want)

        :param start_from_current: boolean. If True: when a process is
          monitored by this handler for the first time, the regex
          matching is started from the position in the stream at the
          time that this output hander starts receiving data. If
          False: when a process is monitored by this handler for the
          first time, the regex matching is started from the beginning
          of the stream.

        """
        if timeout == False: timeout = self.default_expect_timeout
        countdown = Timer(timeout)
        cond = threading.Condition()
        num_found_and_list = [0, {}]
        for p in self.processes: num_found_and_list[1][p] = (None, None)
        def internal_callback(process, stream, re_index, match_object):
            num_found_and_list[0] +=1
            num_found_and_list[1][process] = (re_index, match_object)
            with cond:
                cond.notify_all()
        if self._thread_local_storage.expect_handler == None:
            self._thread_local_storage.expect_handler = ExpectOutputHandler()
        self._thread_local_storage.expect_handler.expect(regexes,
                                                         callback = internal_callback,
                                                         backtrack_size = backtrack_size,
                                                         start_from_current = start_from_current)
        with cond:
            for p in self.processes:
                if stream == STDOUT:
                    p.stdout_handlers.append(self._thread_local_storage.expect_handler)
                else:
                    p.stderr_handlers.append(self._thread_local_storage.expect_handler)
            while (countdown.remaining() == None or countdown.remaining() > 0) and num_found_and_list[0] < len(self.processes):
                non_retrying_intr_cond_wait(cond, countdown.remaining())
        retval = []
        for p in self.processes:
            if num_found_and_list[1][p][0] == None:
                p._notify_expect_fail(regexes)
            retval.append((p, num_found_and_list[1][p][0], num_found_and_list[1][p][1]))
        return retval

def wait_any_actions(actions, timeout = None):
    """Wait for any of the actions given to terminate.

    :param actions: An iterable of `execo.action.Action`.

    :param timeout: Optional timeout in any type supported by
      `execo.time_utils.get_seconds`.

    returns: iterable of `execo.action.Action` which have terminated.
    """
    timeout = get_seconds(timeout)
    if timeout != None:
        end = time.time() + timeout
    with Action._wait_multiple_actions_condition:
        finished = [action for action in actions if action.ended]
        while len(finished) == 0 and (timeout == None or timeout > 0):
            non_retrying_intr_cond_wait(Action._wait_multiple_actions_condition, get_seconds(timeout))
            finished = [action for action in actions if action.ended]
            if timeout != None:
                timeout = end - time.time()
        return finished

def wait_all_actions(actions, timeout = None):
    """Wait for all of the actions given to terminate.

    :param actions: An iterable of `execo.action.Action`.

    :param timeout: Optional timeout in any type supported by
      `execo.time_utils.get_seconds`.

    returns: iterable of `execo.action.Action` which have terminated.
    """
    timeout = get_seconds(timeout)
    if timeout != None:
        end = time.time() + timeout
    with Action._wait_multiple_actions_condition:
        finished = [action for action in actions if action.ended]
        while len(finished) != len(actions) and (timeout == None or timeout > 0):
            non_retrying_intr_cond_wait(Action._wait_multiple_actions_condition, get_seconds(timeout))
            finished = [action for action in actions if action.ended]
            if timeout != None:
                timeout = end - time.time()
        return finished

def filter_bad_hosts(action, hosts):
    """Returns the list of host filtered from any host where any process of the action has failed.

    :param action: The action from which the processes are checked

    :param hosts: The iterable of hosts or host addesses to be
      filtered and returned as a list
    """
    action_hosts_ko = set()
    for p in action.processes:
        if hasattr(p, "host") and not p.ok:
            action_hosts_ko.add(p.host)
    filtered_hosts = []
    for h in hosts:
        if isinstance(h, Host):
            if h not in action_hosts_ko:
                filtered_hosts.append(h)
        else:
            if Host(h) not in action_hosts_ko:
                filtered_hosts.append(h)
    return filtered_hosts

class ActionNotificationProcessLH(ProcessLifecycleHandler):

    def __init__(self, action, total_processes):
        super(ActionNotificationProcessLH, self).__init__()
        self.action = action
        self.total_processes = total_processes
        self.terminated_processes = 0

    def end(self, process):
        self.terminated_processes += 1
        logger.debug("%i/%i processes terminated in %s",
            self.terminated_processes,
            self.total_processes,
            self.action)
        if self.terminated_processes == self.total_processes:
            self.action._notify_terminated()

    def action_reset(self):
        self.terminated_processes = 0

class Remote(Action):

    """Launch a command remotely on several host, with ``ssh`` or a similar remote connection tool.

    One ssh process is launched for each connection.
    """

    def __init__(self, cmd, hosts, connection_params = None, process_args = None, **kwargs):
        """:param cmd: the command to run remotely. Substitions
          described in `execo.substitutions.remote_substitute` will be
          performed.

        :param hosts: iterable of `execo.host.Host` to which to
          connect and run the command.

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.

        :param process_args: Dict of keyword arguments passed to
          instanciated processes.
        """
        self.cmd = cmd
        """The command to run remotely. substitions described in
        `execo.substitutions.remote_substitute` will be performed."""
        if "name" not in kwargs:
            kwargs.update({"name": name_from_cmdline(self.cmd)})
        super(Remote, self).__init__(**kwargs)
        self.connection_params = connection_params
        """A dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        if process_args != None:
            self.process_args = process_args
            """Dict of keyword arguments passed to instanciated processes."""
        else:
            self.process_args = {}
        self.hosts = hosts
        """Iterable of `execo.host.Host` to which to connect and run the command."""
        self._caller_context = get_caller_context(['get_remote'])
        self._init_processes()

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, v):
        self._hosts = get_hosts_list(singleton_to_collection(v))

    def _args(self):
        return [ repr(self.cmd),
                 repr(self.hosts) ] + Action._args(self) + Remote._kwargs(self)

    def _kwargs(self):
        kwargs = []
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        if len(self.process_args) > 0: kwargs.append("process_args=%r" % (self.process_args,))
        return kwargs

    def _infos(self):
        return Action._infos(self)

    def _init_processes(self):
        self.processes = []
        processlh = ActionNotificationProcessLH(self, len(self.hosts))
        for (index, host) in enumerate(self.hosts):
            p = SshProcess(remote_substitute(self.cmd, self.hosts, index, self._caller_context),
                           host = host,
                           connection_params = self.connection_params,
                           **self.process_args)
            p.lifecycle_handlers.append(processlh)
            self.processes.append(p)

    def start(self):
        retval = super(Remote, self).start()
        if len(self.processes) == 0:
            logger.debug("%s contains 0 processes -> immediately terminated", self)
            self._notify_terminated()
        else:
            for process in self.processes:
                process.start()
        return retval

    def kill(self):
        retval = super(Remote, self).kill()
        for process in self.processes:
            if process.running:
                process.kill()
        return retval

    def write(self, s):
        """Write on the Remote processes standard inputs

        Allows Remote instances to behave as file-like objects. You
        can for example print to a Remote.
        """
        for process in self.processes:
            process.write(s)

class _TaktukRemoteOutputHandler(ProcessOutputHandler):

    """Parse taktuk output."""

    def __init__(self, taktukaction):
        super(_TaktukRemoteOutputHandler, self).__init__()
        self.taktukaction = taktukaction

    def _log_unexpected_output(self, string):
        logger.critical("%s: Taktuk unexpected output parsing. Please report this message. Line received:", self.__class__.__name__)
        logger.critical(self._describe_taktuk_output(string))

    def _describe_taktuk_output(self, string):
        s = "%s: " % (self.__class__.__name__,)
        try:
            if len(string) > 0:
                header = ord(string[0])
                (position, _, line) = string[2:].partition(" # ")
                position = int(position)
                if position == 0:
                    host_address = "localhost"
                else:
                    host_address = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]].host.address
                if header in (65, 66, 67, 70, 71, 72):
                    if header == 65: t = "stdout"
                    elif header == 66: t = "stderr"
                    elif header == 67: t = "status"
                    elif header == 70: t = "info"
                    elif header == 71: t = "taktuk"
                    elif header == 72: t = "message"
                    s += "%s - host = %s - line = %s" % (t, host_address, line[:-1])
                elif header in (68, 69):
                    (peer_position, _, line) = line.partition(" # ")
                    peer_host_address = None
                    try:
                        peer_position = int(peer_position)
                        if peer_position == 0:
                            peer_host_address = "localhost"
                        else:
                            peer_host_address = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]].host.address
                    except ValueError:
                        pass
                    if header == 68:
                        s += "connector - host = %s - peer = %s - line = %s" % (host_address, peer_host_address, line[:-1])
                    elif header == 69:
                        (_, _, state_msg) = line.partition(" # ")
                        s += "state - host = %s - peer = %s - state = %s" % (host_address, peer_host_address, state_msg[:-1])
                elif header == 73:
                    (taktuktype, _, line) = line.partition(" # ")
                    s += "message type = %s - host = %s - line = %s" % (taktuktype, host_address, line[:-1])
                else:
                    s += "unexpected string = %s" % (string[:-1])
            else:
                s += "empty string"
            return s
        except Exception as e: #IGNORE:W0703
            logger.critical("%s: Unexpected exception %s while parsing taktuk output. Please report this message.", self.__class__.__name__, e)
            logger.critical("line received = %s", string.rstrip('\n'))
            return s

    def read_line(self, process, stream, string):
        # my taktuk output protocol:
        #  stream    format                                                    header normal?
        #  output    "A $position # $line"                                     65     YES
        #  error     "B $position # $line"                                     66     YES
        #  status    "C $position # $line"                                     67     YES
        #  connector "D $position # $peer_position # $line"                    68     YES
        #  state     "E $position # $peer_position # $line # event_msg($line)" 69     YES
        #  info      "F $position # $line"                                     70     NO
        #  taktuk    "G $position # $line"                                     71     SOMETIMES
        #  message   "H $position # $line"                                     72     NO
        #  default   "I $position # $type # $line"                             73     NO
        try:
            if len(string) > 0:
                #logger.debug("taktuk: %s", self._describe_taktuk_output(string))
                header = ord(string[0])
                (position, _, line) = string[2:].partition(" # ")
                position = int(position)
                if header >= 65 and header <= 67: # stdout, stderr, status
                    if position == 0:
                        self._log_unexpected_output(string)
                        return
                    else:
                        process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]]
                        if header == 65: # stdout
                            process._handle_stdout(line, False, False)
                        elif header == 66: # stderr
                            process._handle_stderr(line, False, False)
                        else: # 67: status
                            process._set_terminated(exit_code = int(line))
                elif header in (68, 69): # connector, state
                    (peer_position, _, line) = line.partition(" # ")
                    if header == 68: # connector
                        peer_position = int(peer_position)
                        process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                        process._handle_stderr(line, False, False)
                    else: # state
                        (state_code, _, _) = line.partition(" # ")
                        state_code = int(state_code)
                        if state_code == 6 or state_code == 7: # command started or remote command exec failed
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]]
                            if state_code == 6: # command started
                                process.start()
                            else: # 7: remote command exec failed
                                process._set_terminated(error = True, error_reason = "taktuk remote command execution failed")
                        elif state_code == 3 or state_code == 5: # connection failed or lost
                            peer_position = int(peer_position)
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                            if state_code == 3: # connection failed
                                process._set_terminated(error = True, error_reason = "taktuk connection failed")
                            else: # 5: connection lost
                                process._set_terminated(error = True, error_reason = "taktuk connection lost")
                        elif state_code in (0, 1, 2, 4, 8, 19):
                            pass
                        else:
                            self._log_unexpected_output(string)
                else:
                    self._log_unexpected_output(string)
        except Exception as e: #IGNORE:W0703
            logger.critical("%s: Unexpected exception %s while parsing taktuk output. Please report this message.", self.__class__.__name__, e)
            logger.critical("line received = %s", string.rstrip('\n'))
            raise

class _TaktukLH(ProcessLifecycleHandler):

    """Notify `execo.process.TaktukProcess` of their real taktuk process lifecyle."""

    def __init__(self, taktukremote):
        super(_TaktukLH, self).__init__()
        self.taktukremote = taktukremote

    def end(self, process):
        error = None
        error_reason = None
        timeouted = None
        forced_kill = None
        if process.error == True:
            error = True
        if process.error_reason != None:
            error_reason = "error of taktuk process: %s" % process.error_reason
        if process.timeouted == True:
            timeouted = True
        if process.forced_kill == True:
            forced_kill = True
        for taktukprocess in self.taktukremote.processes:
            if not taktukprocess.ended:
                taktukprocess._set_terminated(error = error,
                                              error_reason = error_reason,
                                              timeouted = timeouted,
                                              forced_kill = forced_kill)

def _escape_brackets_in_taktuk_options(s):
    for c in [ '[', ']' ]:
        s = s.replace( c, '\\' + c )
    return s

class TaktukRemote(Action):

    """Launch a command remotely on several host, with ``taktuk``.

    One taktuk instance is ran, which itself connects to hosts through
    an ``ssh`` tree.

    Behavior should be identical to `execo.action.Remote`. Current
    limitation are:

    - we can provide per-host user with taktuk, but we cannot provide
      per-host port or keyfile, so a check is made that all hosts and
      connection_params have the same port / keyfile (or None). If not,
      an exception is raised during initialization.

    - remote processes are not killed when killing the
      TaktukRemote. See 'hanged commands' in
      http://taktuk.gforge.inria.fr/taktuk.html#bugs. With ssh the
      workaround is to pass options -tt but passing these options to
      taktuk connector causes immediate closing of the connector upon
      connection.
    """

    def __init__(self, cmd, hosts, connection_params = None, process_args = None, **kwargs):
        """
        :param cmd: the command to run remotely. substitions
          described in `execo.substitutions.remote_substitute` will be
          performed.

        :param hosts: iterable of `execo.host.Host` to which to
          connect and run the command.

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.cmd = cmd
        """The command to run remotely. substitions described in
        `execo.substitutions.remote_substitute` will be performed."""
        if "name" not in kwargs:
            kwargs.update({"name": name_from_cmdline(self.cmd)})
        super(TaktukRemote, self).__init__(**kwargs)
        self.connection_params = connection_params
        """A dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        if process_args != None:
            self.process_args = process_args
            """Dict of keyword arguments passed to instanciated processes."""
        else:
            self.process_args = {}
        self.hosts = hosts
        """Iterable of `execo.host.Host` to which to connect and run the command."""
        self._caller_context = get_caller_context(['get_remote'])
        self._taktuk_stdout_output_handler = _TaktukRemoteOutputHandler(self)
        self._taktuk_stderr_output_handler = self._taktuk_stdout_output_handler
        self._init_processes()

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, v):
        self._hosts = get_hosts_list(singleton_to_collection(v))

    def _args(self):
        return [ repr(self.cmd),
                 repr(self.hosts) ] + Action._args(self) + TaktukRemote._kwargs(self)

    def _kwargs(self):
        kwargs = []
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        if len(self.process_args) > 0: kwargs.append("process_args=%r" % (self.process_args,))
        return kwargs

    def _gen_taktukprocesses(self):
        processlh = ActionNotificationProcessLH(self, len(self.hosts))
        for (index, host) in enumerate(self.hosts):
            p = TaktukProcess(remote_substitute(self.cmd, self.hosts, index, self._caller_context),
                              host = host,
                              **self.process_args)
            p.lifecycle_handlers.append(processlh)
            p._taktuk_remote = self
            self.processes.append(p)

    def _gen_taktuk_commands(self, hosts_with_explicit_user):
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h not in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-m", get_rewritten_host_address(host.address, self.connection_params), "-[", "exec", "[", _escape_brackets_in_taktuk_options(self.processes[index].cmd), "]", "-]",)
            self._taktuk_hosts_order.append(index)
            self.processes[index]._taktuk_index = index
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-l", host.user, "-m", get_rewritten_host_address(host.address, self.connection_params), "-[", "exec", "[", _escape_brackets_in_taktuk_options(self.processes[index].cmd), "]", "-]",)
            self._taktuk_hosts_order.append(index)
            self.processes[index]._taktuk_index = index

    def _init_processes(self):
        # taktuk code common to TaktukRemote and subclasses TaktukGet
        # TaktukPut
        self.processes = []
        self._taktuk_hosts_order = []
        self._taktuk_commands = ()
        self._taktuk = None
        if len(self.hosts) > 0:
            # we can provide per-host user with taktuk, but we cannot
            # provide per-host port or keyfile, so check that all hosts
            # and connection_params have the same port / keyfile (or None)
            actual_connection_params = make_connection_params(self.connection_params)
            check_default_port = actual_connection_params['port']
            check_default_keyfile = actual_connection_params['keyfile']
            check_keyfiles = set()
            check_ports = set()
            hosts_with_explicit_user = set()
            for host in self.hosts:
                if host.user != None:
                    hosts_with_explicit_user.add(host)
                if host.keyfile != None:
                    check_keyfiles.add(host.keyfile)
                else:
                    check_keyfiles.add(check_default_keyfile)
                if host.port != None:
                    check_ports.add(host.port)
                else:
                    check_ports.add(check_default_port)
            if len(check_keyfiles) > 1 or len(check_ports) > 1:
                raise ValueError("unable to provide more than one keyfile / port for taktuk remote connection")
            global_keyfile = None
            global_port = None
            if len(check_keyfiles) == 1:
                global_keyfile = list(check_keyfiles)[0]
            if len(check_ports) == 1:
                global_port = list(check_ports)[0]
            self._gen_taktukprocesses()
            self._gen_taktuk_commands(hosts_with_explicit_user)
            #self._taktuk_commands += ("quit",)
            #handler = _TaktukRemoteOutputHandler(self)
            taktuk_options_filehandle, taktuk_options_filename = tempfile.mkstemp(prefix = 'tmp_execo_taktuk_')
            self._taktuk_commands = " ".join(self._taktuk_commands)
            os.write(taktuk_options_filehandle, self._taktuk_commands + "\n")
            os.close(taktuk_options_filehandle)
            logger.debug("generated taktuk tmp cmd file %s with content:\n%s", taktuk_options_filename, self._taktuk_commands)
            real_taktuk_cmdline = (actual_connection_params['taktuk'],)
            real_taktuk_cmdline += ("-E", "!")
            real_taktuk_cmdline += ("-o", 'output="A $position # $line\\n"',
                                    "-o", 'error="B $position # $line\\n"',
                                    "-o", 'status="C $position # $line\\n"',
                                    "-o", 'connector="D $position # $peer_position # $line\\n"',
                                    "-o", 'state="E $position # $peer_position # $line # ".event_msg($line)."\\n"',
                                    "-o", 'info="F $position # $line\\n"',
                                    "-o", 'taktuk="G $position # $line\\n"',
                                    "-o", 'message="H $position # $line\\n"',
                                    "-o", 'default="I $position # $type > $line\\n"')
            real_taktuk_cmdline += actual_connection_params['taktuk_options']
            real_taktuk_cmdline += ("-c", " ".join(
                get_taktuk_connector_command(keyfile = global_keyfile,
                                             port = global_port,
                                             connection_params = self.connection_params)))
            real_taktuk_cmdline += ("-F", taktuk_options_filename)
            real_taktuk_cmdline = " ".join([pipes.quote(arg) for arg in real_taktuk_cmdline])
            real_taktuk_cmdline += " && rm -f " + taktuk_options_filename
            self._taktuk = Process(real_taktuk_cmdline)
            #self._taktuk.close_stdin = False
            self._taktuk.shell = True
            #self._taktuk.default_stdout_handler = False
            #self._taktuk.default_stderr_handler = False
            self._taktuk.stdout_handlers.append(self._taktuk_stdout_output_handler)
            #self._taktuk.stderr_handlers.append(self._taktuk_stderr_output_handler)
            self._taktuk.lifecycle_handlers.append(_TaktukLH(self))

    def start(self):
        retval = super(TaktukRemote, self).start()
        if len(self.processes) == 0:
            logger.debug("%s contains 0 processes -> immediately terminated", self)
            self._notify_terminated()
        else:
            self._taktuk.start()
        return retval

    def kill(self):
        retval = super(TaktukRemote, self).kill()
        if self._taktuk:
            for process in self.processes:
                if process.running:
                    process.killed = True
            self._taktuk.write("broadcast kill\n\n")
            self._taktuk.write("quit\n\n")
        return retval

    def wait(self, timeout = None):
        retval = super(TaktukRemote, self).wait(timeout)
        if self._taktuk:
            self._taktuk.wait()
        return retval

    def _notify_terminated(self):
        try:
            self._taktuk.write("quit\n\n")
        except:
            pass
        super(TaktukRemote, self)._notify_terminated()

class Put(Remote):

    """Copy local files to several remote host, with ``scp`` or a similar connection tool."""

    def __init__(self, hosts, local_files, remote_location = ".", connection_params = None, **kwargs):
        """
        :param hosts: iterable of `execo.host.Host` onto which to copy
          the files.

        :param local_files: an iterable of string of file
          paths. substitions described in
          `execo.substitutions.remote_substitute` will be performed.

        :param remote_location: the directory on the remote hosts were
          the files will be copied. substitions described in
          `execo.substitutions.remote_substitute` will be performed.

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.hosts = hosts
        """Iterable of `execo.host.Host` onto which to copy the files."""
        if "name" not in kwargs:
            kwargs.update({"name": "%s to %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(Remote, self).__init__(**kwargs)
        self.local_files = local_files
        """An iterable of string of file paths. substitions described in
        `execo.substitutions.remote_substitute` will be performed."""
        self.remote_location = remote_location
        """The directory on the remote hosts were the files will be
        copied. substitions described in `execo.substitutions.remote_substitute`
        will be performed."""
        self.connection_params = connection_params
        """A dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        self._caller_context = get_caller_context(['get_fileput'])
        self._init_processes()

    def _args(self):
        return [ repr(self.hosts),
                 repr(self.local_files) ] + Action._args(self) + Put._kwargs(self)

    def _kwargs(self):
        kwargs = []
        kwargs.append("remote_location=%r" % (self.remote_location,))
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        return kwargs

    def _init_processes(self):
        self.processes = []
        if len(self.local_files) > 0:
            processlh = ActionNotificationProcessLH(self, len(self.hosts))
            for (index, host) in enumerate(self.hosts):
                real_command = list(get_scp_command(host.user, host.keyfile, host.port, self.connection_params)) + [ remote_substitute(local_file, self.hosts, index, self._caller_context) for local_file in self.local_files ] + ["%s:%s" % (get_rewritten_host_address(host.address, self.connection_params), remote_substitute(self.remote_location, self.hosts, index, self._caller_context)),]
                real_command = ' '.join(real_command)
                p = Process(real_command)
                p.shell = True
                p.lifecycle_handlers.append(processlh)
                p.host = host
                self.processes.append(p)

class Get(Remote):

    """Copy remote files from several remote host to a local directory, with ``scp`` or a similar connection tool."""

    def __init__(self, hosts, remote_files, local_location = ".", connection_params = None, **kwargs):
        """
        :param hosts: iterable of `execo.host.Host` from which to get
          the files.

        :param remote_files: an iterable of string of file
          paths. substitions described in
          `execo.substitutions.remote_substitute` will be performed.

        :param local_location: the local directory were the files will
          be copied. substitions described in
          `execo.substitutions.remote_substitute` will be performed.

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.hosts = hosts
        """Iterable of `execo.host.Host` from which to get the files."""
        if "name" not in kwargs:
            kwargs.update({"name": "%s from %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(Remote, self).__init__(**kwargs)
        self.remote_files = remote_files
        """Iterable of string of file paths. substitions described in
        `execo.substitutions.remote_substitute` will be performed."""
        self.local_location = local_location
        """The local directory were the files will be copied. substitions described
        in `execo.substitutions.remote_substitute` will be performed."""
        self.connection_params = connection_params
        """Dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        self._caller_context = get_caller_context(['get_fileget'])
        self._init_processes()

    def _args(self):
        return [ repr(self.hosts),
                 repr(self.remote_files) ] + Action._args(self) + Get._kwargs(self)

    def _kwargs(self):
        kwargs = []
        kwargs.append("local_location=%r" % (self.local_location,))
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        return kwargs

    def _init_processes(self):
        self.processes = []
        if len(self.remote_files) > 0:
            processlh = ActionNotificationProcessLH(self, len(self.hosts))
            for (index, host) in enumerate(self.hosts):
                remote_specs = ()
                for path in self.remote_files:
                    remote_specs += ("%s:%s" % (get_rewritten_host_address(host.address, self.connection_params), remote_substitute(path, self.hosts, index, self._caller_context)),)
                real_command = get_scp_command(host.user, host.keyfile, host.port, self.connection_params) + remote_specs + (remote_substitute(self.local_location, self.hosts, index, self._caller_context),)
                real_command = ' '.join(real_command)
                p = Process(real_command)
                p.shell = True
                p.lifecycle_handlers.append(processlh)
                p.host = host
                self.processes.append(p)

class _TaktukPutOutputHandler(_TaktukRemoteOutputHandler):

    """Parse taktuk output."""

    def _update_taktukprocess_end_state(self, process):
        if process._num_transfers_started > 0 and not process.started:
            process.start()
        if process._num_transfers_failed + process._num_transfers_terminated >= len(self.taktukaction.local_files):
            if process._num_transfers_failed > 0:
                process._set_terminated(error = True, error_reason = "taktuk file reception failed")
            else:
                process._set_terminated(exit_code = 0)

    def read_line(self, process, stream, string):
        try:
            if len(string) > 0:
                header = ord(string[0])
                (position, _, line) = string[2:].partition(" # ")
                position = int(position)
                if header in (68, 69): # connector, state
                    (peer_position, _, line) = line.partition(" # ")
                    if header == 68: # connector
                        peer_position = int(peer_position)
                        process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                        process._handle_stderr(line, False, False)
                    else: # state
                        (state_code, _, _) = line.partition(" # ")
                        state_code = int(state_code)
                        if state_code in (13, 14, 15): # file reception started, failed, terminated
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]]
                            if state_code == 13: # file reception started
                                process._num_transfers_started += 1
                            elif state_code == 14: # file reception failed
                                process._num_transfers_failed += 1
                            else: # 15: file reception terminated
                                process._num_transfers_terminated += 1
                            self._update_taktukprocess_end_state(process)
                        elif state_code == 16: # file send failed
                            if position > 0:
                                processes = [ self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]] ]
                            else:
                                processes = self.taktukaction.processes
                            for process in processes:
                                process._num_transfers_started += 1
                                process._num_transfers_failed += 1
                                self._update_taktukprocess_end_state(process)
                        elif state_code == 3 or state_code == 5: # connection failed or lost
                            peer_position = int(peer_position)
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                            if state_code == 3: # connection failed
                                process._set_terminated(error = True, error_reason = "taktuk connection failed")
                            else: # 5: connection lost
                                process._set_terminated(error = True, error_reason = "taktuk connection lost")
                        elif state_code in (0, 1, 2, 4):
                            pass
                        else:
                            self._log_unexpected_output(string)
                elif header == 71: # to ignore "Error No such file or directory". Don't know if it can occur in other situations
                    if position > 0:
                        processes = [ self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]] ]
                    else:
                        processes = self.taktukaction.processes
                    for process in processes:
                        process._handle_stderr(line, False, False)
                else:
                    self._log_unexpected_output(string)
        except Exception as e: #IGNORE:W0703
            logger.critical("%s: Unexpected exception %s while parsing taktuk output. Please report this message.", self.__class__.__name__, e)
            logger.critical("line received = %s", string.rstrip('\n'))

class TaktukPut(TaktukRemote):

    """Copy local files to several remote host, with ``taktuk``."""

    def __init__(self, hosts, local_files, remote_location = ".", connection_params = None, **kwargs):
        """
        :param hosts: iterable of `execo.host.Host` onto which to copy
          the files.

        :param local_files: an iterable of string of file
          paths. substitions described in
          `execo.substitutions.remote_substitute` will not be
          performed, but taktuk substitutions can be used (see
          http://taktuk.gforge.inria.fr/taktuk.html#item_put__2a_src__2a__2a_dest__2a)

        :param remote_location: the directory on the remote hosts were
          the files will be copied. substitions described in
          `execo.substitutions.remote_substitute` will not be
          performed, but taktuk substitutions can be used (see
          http://taktuk.gforge.inria.fr/taktuk.html#item_put__2a_src__2a__2a_dest__2a)

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.hosts = hosts
        """Iterable of `execo.host.Host` onto which to copy the files."""
        if "name" not in kwargs:
            kwargs.update({"name": "%s to %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(TaktukRemote, self).__init__(**kwargs)
        self.local_files = local_files
        """Iterable of string of file paths. substitions described in
        `execo.substitutions.remote_substitute` will not be performed,
        but taktuk substitutions can be used (see
        http://taktuk.gforge.inria.fr/taktuk.html#item_put__2a_src__2a__2a_dest__2a)"""
        self.remote_location = remote_location
        """The directory on the remote hosts were the files will be
        copied. substitions described in `execo.substitutions.remote_substitute`
        will not be performed, but taktuk substitutions can be used (see
        http://taktuk.gforge.inria.fr/taktuk.html#item_put__2a_src__2a__2a_dest__2a)"""
        self.connection_params = connection_params
        """Dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        self.process_args = {}
        self._caller_context = get_caller_context(['get_fileput'])
        self._taktuk_stdout_output_handler = _TaktukPutOutputHandler(self)
        self._taktuk_stderr_output_handler = self._taktuk_stdout_output_handler
        self._init_processes()

    def _args(self):
        return [ repr(self.hosts),
                 repr(self.local_files) ] + Action._args(self) + TaktukPut._kwargs(self)

    def _kwargs(self):
        kwargs = []
        kwargs.append("remote_location=%r" % (self.remote_location,))
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        return kwargs

    def _gen_taktukprocesses(self):
        processlh = ActionNotificationProcessLH(self, len(self.hosts))
        for (index, host) in enumerate(self.hosts):
            process = TaktukProcess("",
                                    host = host,
                                    **self.process_args)
            process.lifecycle_handlers.append(processlh)
            process._num_transfers_started = 0
            process._num_transfers_terminated = 0
            process._num_transfers_failed = 0
            self.processes.append(process)

    def _gen_taktuk_commands(self, hosts_with_explicit_user):
        self._taktuk_hosts_order = []
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h not in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-m", get_rewritten_host_address(host.address, self.connection_params))
            self._taktuk_hosts_order.append(index)
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-l", host.user, "-m", get_rewritten_host_address(host.address, self.connection_params))
            self._taktuk_hosts_order.append(index)
        for src in self.local_files:
            self._taktuk_commands += ("broadcast", "put", "[", src, "]", "[", self.remote_location, "]", ";")

class _TaktukGetOutputHandler(_TaktukRemoteOutputHandler):

    """Parse taktuk output."""

    def _update_taktukprocess_end_state(self, process):
        if process._num_transfers_started > 0 and not process.started:
            process.start()
        if process._num_transfers_failed + process._num_transfers_terminated >= len(self.taktukaction.remote_files):
            if process._num_transfers_failed > 0:
                process._set_terminated(error = True, error_reason = "taktuk file reception failed")
            else:
                process._set_terminated(exit_code = 0)

    def read_line(self, process, stream, string):
        try:
            if len(string) > 0:
                header = ord(string[0])
                (position, _, line) = string[2:].partition(" # ")
                position = int(position)
                if header in (68, 69): # connector, state
                    (peer_position, _, line) = line.partition(" # ")
                    if header == 68: # connector
                        peer_position = int(peer_position)
                        process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                        process._handle_stderr(line, False, False)
                    else: # state
                        (state_code, _, _) = line.partition(" # ")
                        state_code = int(state_code)
                        if state_code in (13, 14, 15): # file reception started, failed, terminated
                            peer_position = int(peer_position)
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                            if state_code == 13: # file reception started
                                process._num_transfers_started += 1
                            elif state_code == 14: # file reception failed
                                process._num_transfers_failed += 1
                            else: # 15: file reception terminated
                                process._num_transfers_terminated += 1
                            self._update_taktukprocess_end_state(process)
                        elif state_code == 16: # file send failed
                            if position > 0:
                                processes = [ self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]] ]
                            else:
                                processes = self.taktukaction.processes
                            for process in processes:
                                process._num_transfers_started += 1
                                process._num_transfers_failed += 1
                                self._update_taktukprocess_end_state(process)
                        elif state_code == 3 or state_code == 5: # connection failed or lost
                            peer_position = int(peer_position)
                            process = self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[peer_position-1]]
                            if state_code == 3: # connection failed
                                process._set_terminated(error = True, error_reason = "taktuk connection failed")
                            else: # 5: connection lost
                                process._set_terminated(error = True, error_reason = "taktuk connection lost")
                        elif state_code in (0, 1, 2, 4, 19):
                            pass
                        else:
                            self._log_unexpected_output(string)
                elif header == 71: # to ignore "Error No such file or directory". Don't know if it can occur in other situations
                    if position > 0:
                        processes = [ self.taktukaction.processes[self.taktukaction._taktuk_hosts_order[position-1]] ]
                    else:
                        processes = self.taktukaction.processes
                    for process in processes:
                        process._handle_stderr(line, False, False)
                else:
                    self._log_unexpected_output(string)
        except Exception as e: #IGNORE:W0703
            logger.critical("%s: Unexpected exception %s while parsing taktuk output. Please report this message.", self.__class__.__name__, e)
            logger.critical("line received = %s", string.rstrip('\n'))

class TaktukGet(TaktukRemote):

    """Copy remote files from several remote host to a local directory, with ``taktuk``."""

    def __init__(self, hosts, remote_files, local_location = ".", connection_params = None, **kwargs):
        """
        :param hosts: iterable of `execo.host.Host` from which to get
          the files.

        :param remote_files: an iterable of string of file
          paths. Substitions described in
          `execo.substitutions.remote_substitute` will not be
          performed, but taktuk substitutions can be used (see
          http://taktuk.gforge.inria.fr/taktuk.html#item_get__2a_src__2a__2a_dest__2a)

        :param local_location: the local directory were the files will
          be copied. Substitions described in
          `execo.substitutions.remote_substitute` will not be
          performed, but taktuk substitutions can be used (see
          http://taktuk.gforge.inria.fr/taktuk.html#item_get__2a_src__2a__2a_dest__2a)

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.hosts = hosts
        """Iterable of `execo.host.Host` from which to get the files."""
        if "name" not in kwargs:
            kwargs.update({"name": "%s from %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(TaktukRemote, self).__init__(**kwargs)
        self.remote_files = remote_files
        """Iterable of string of file paths. Substitions described in
        `execo.substitutions.remote_substitute` will not be performed, but
        taktuk substitutions can be used (see
        http://taktuk.gforge.inria.fr/taktuk.html#item_get__2a_src__2a__2a_dest__2a)"""
        self.local_location = local_location
        """The local directory were the files will be copied. Substitions described
        in `execo.substitutions.remote_substitute` will not be performed, but
        taktuk substitutions can be used (see
        http://taktuk.gforge.inria.fr/taktuk.html#item_get__2a_src__2a__2a_dest__2a)"""
        self.connection_params = connection_params
        """Dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        self.process_args = {}
        self._caller_context = get_caller_context(['get_fileget'])
        self._taktuk_stdout_output_handler = _TaktukGetOutputHandler(self)
        self._taktuk_stderr_output_handler = self._taktuk_stdout_output_handler
        self._init_processes()

    def _args(self):
        return [ repr(self.hosts),
                 repr(self.remote_files) ] + Action._args(self) + TaktukGet._kwargs(self)

    def _kwargs(self):
        kwargs = []
        kwargs.append("local_location=%r" % (self.local_location,))
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        return kwargs

    def _gen_taktukprocesses(self):
        processlh = ActionNotificationProcessLH(self, len(self.hosts))
        for (index, host) in enumerate(self.hosts):
            process = TaktukProcess("",
                                    host = host,
                                    **self.process_args)
            process.lifecycle_handlers.append(processlh)
            process._num_transfers_started = 0
            process._num_transfers_terminated = 0
            process._num_transfers_failed = 0
            self.processes.append(process)

    def _gen_taktuk_commands(self, hosts_with_explicit_user):
        self._taktuk_hosts_order = []
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h not in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-m", get_rewritten_host_address(host.address, self.connection_params))
            self._taktuk_hosts_order.append(index)
        for (index, host) in [ (idx, h) for (idx, h) in enumerate(self.hosts) if h in hosts_with_explicit_user ]:
            self._taktuk_commands += ("-l", host.user, "-m", get_rewritten_host_address(host.address, self.connection_params))
            self._taktuk_hosts_order.append(index)
        for src in self.remote_files:
            self._taktuk_commands += ("broadcast", "get", "[", src, "]", "[", self.local_location, "]", ";")

class Local(Action):

    """Launch a command localy."""

    def __init__(self, cmd, process_args = None, **kwargs):
        """:param cmd: the command to run.

        :param process_args: Dict of keyword arguments passed to
          instanciated processes.
        """
        self.cmd = cmd
        """the command to run"""
        if "name" not in kwargs:
            kwargs.update({"name": name_from_cmdline(self.cmd)})
        super(Local, self).__init__(**kwargs)
        if process_args != None:
            self.process_args = process_args
            """Dict of keyword arguments passed to instanciated processes."""
        else:
            self.process_args = {}
        self._init_processes()

    def _args(self):
        return [ repr(self.cmd) ] + Action._args(self) + Local._kwargs(self)

    def _kwargs(self):
        return []

    def _init_processes(self):
        self.processes = []
        p = Process(self.cmd, **self.process_args)
        processlh = ActionNotificationProcessLH(self, 1)
        p.lifecycle_handlers.append(processlh)
        self.processes.append(p)

    def start(self):
        retval = super(Local, self).start()
        [ p.start() for p in self.processes ]
        return retval

    def kill(self):
        retval = super(Local, self).kill()
        [ p.kill() for p in self.processes if p.running ]
        return retval

class ParallelSubActionLH(ActionLifecycleHandler):

    def __init__(self, parallel_action):
        super(ParallelSubActionLH, self).__init__()
        self.parallel_action = parallel_action
        self._num_terminated = 0

    def reset(self, action):
        self._num_terminated = 0

    def end(self, action):
        self._num_terminated += 1
        logger.debug("%i/%i subactions terminated in %s",
            self._num_terminated,
            len(self.parallel_action.actions),
            self.parallel_action)
        if self._num_terminated == len(self.parallel_action.actions):
            self.parallel_action._notify_terminated()

class ParallelActions(Action):

    """An `execo.action.Action` running several sub-Actions in parallel.

    Will start, kill, wait, run every action in parallel.
    """

    def __init__(self, actions, **kwargs):
        self.actions = actions
        if "name" not in kwargs:
            kwargs.update({"name": "%s %i actions" % (self.__class__.__name__, len(self.actions))})
        super(ParallelActions, self).__init__(**kwargs)
        self.hide_subactions = False
        """Wether to hide sub actions in stats."""
        self._init_actions()

    def _args(self):
        return [ repr(self.actions) ] + Action._args(self) + ParallelActions._kwargs(self)

    def _kwargs(self):
        return []

    def _init_actions(self):
        subactionslh = ParallelSubActionLH(self)
        for action in self.actions:
            action.lifecycle_handlers = [ lh for lh in action.lifecycle_handlers if not isinstance(lh, ParallelSubActionLH) ]
            action.lifecycle_handlers.append(subactionslh)

    def start(self):
        retval = super(ParallelActions, self).start()
        if len(self.actions) == 0:
            logger.debug("%s contains 0 actions -> immediately terminated", self)
            self._notify_terminated()
        else:
            for action in self.actions:
                action.start()
        return retval

    def kill(self):
        retval = super(ParallelActions, self).kill()
        for action in self.actions:
            action.kill()
        return retval

    def reset(self):
        retval = super(ParallelActions, self).reset()
        for action in self.actions:
            action.reset()
        self._init_actions()
        return retval

    @property
    def processes(self):
        p = []
        for action in self.actions:
            p.extend(action.processes)
        return p

    @processes.setter
    def processes(self, v):
        pass

    def stats(self):
        stats = Report.empty_stats()
        stats['name'] = self.name
        stats['sub_stats'] = [action.stats() for action in self.actions]
        s = Report.aggregate_stats(stats)
        if self.hide_subactions:
            s['sub_stats'] = []
        return s

class SequentialSubActionLH(ActionLifecycleHandler):

    def __init__(self, sequential_action):
        super(SequentialSubActionLH, self).__init__()
        self.sequential_action = sequential_action
        self._num_terminated = 0

    def reset(self, action):
        self._num_terminated = 0

    def end(self, action):
        self._num_terminated += 1
        logger.debug("%i/%i subactions terminated in %s",
            self._num_terminated,
            len(self.sequential_action.actions),
            self.sequential_action)
        if self._num_terminated < len(self.sequential_action.actions):
            self.sequential_action.actions[self._num_terminated].start()
        else:
            self.sequential_action._notify_terminated()

class SequentialActions(Action):

    """An `execo.action.Action` running several sub-actions sequentially.

    Will start, kill, wait, run every Action sequentially.
    """

    def __init__(self, actions, **kwargs):
        self.actions = actions
        if "name" not in kwargs:
            kwargs.update({"name": "%s %i actions" % (self.__class__.__name__, len(self.actions))})
        super(SequentialActions, self).__init__(**kwargs)
        self.hide_subactions = False
        """Wether to hide sub actions in stats."""
        self._init_actions()

    def _args(self):
        return [ repr(self.actions) ] + Action._args(self) + SequentialActions._kwargs(self)

    def _kwargs(self):
        return []

    def _init_actions(self):
        subactionslh = SequentialSubActionLH(self)
        for action in self.actions:
            action.lifecycle_handlers = [ lh for lh in action.lifecycle_handlers if not isinstance(lh, SequentialSubActionLH) ]
            action.lifecycle_handlers.append(subactionslh)

    def start(self):
        retval = super(SequentialActions, self).start()
        if len(self.actions) == 0:
            logger.debug("%s contains 0 actions -> immediately terminated", self)
            self._notify_terminated()
        else:
            self.actions[0].start()
        return retval

    def kill(self):
        retval = super(SequentialActions, self).kill()
        for action in self.actions:
            action.kill()
        return retval

    def reset(self):
        retval = super(SequentialActions, self).reset()
        for action in self.actions:
            action.reset()
        self._init_actions()
        return retval

    @property
    def processes(self):
        p = []
        for action in self.actions:
            p.extend(action.processes)
        return p

    @processes.setter
    def processes(self, v):
        pass

    def stats(self):
        stats = Report.empty_stats()
        stats['name'] = self.name
        stats['sub_stats'] = [action.stats() for action in self.actions]
        s = Report.aggregate_stats(stats)
        if self.hide_subactions:
            s['sub_stats'] = []
        return s

# class _ChainPutActionHostFilteringLH(ActionLifecycleHandler):

#     def __init__(self, chainput, filtered_actions):
#         super(_ChainPutActionHostFilteringLH, self).__init__()
#         self.chainput = chainput
#         self.filtered_actions = filtered_actions

#     def end(self, action):
#         bad_hosts = [ p.host for p in
#                       action.processes
#                       if not p.ok ]
#         logger.debug(
#             "_ChainPutActionHostFilteringLH: action %s finished. bad hosts = %s" % (
#                 action, bad_hosts))
#         # partie qui pose souci
#         # for a in self.filtered_actions:
#         #     logger.debug(
#         #         "removing hosts %s from action %s" % (
#         #             set(a.hosts).intersection(bad_hosts), a))
#         #     a.hosts = list(set(a.hosts).difference(bad_hosts))
#         #     a.reset()
#         # fin partie qui pose souci
#         self.chainput.bad_hosts.update(bad_hosts)
#         self.chainput.good_hosts.difference(bad_hosts)
#         logger.debug(
#             "%s:\ngood_hosts = %s\nbad_hosts = %s" % (
#                 self.chainput,
#                 self.chainput.good_hosts,
#                 self.chainput.bad_hosts))

class _ChainPutCopyTaktukProcessLH(ProcessLifecycleHandler):

    def __init__(self, num_processes, action_to_start):
        self.num_processes = num_processes
        self.started_processes = 0
        self.action_to_start = action_to_start

    def start(self, process):
        self.started_processes += 1
        if self.started_processes == self.num_processes:
            self.action_to_start.start()

class _ChainPutCopy(ParallelActions):

    def __init__(self, local, remote):
        self._local = local
        self._remote = remote
        super(_ChainPutCopy, self).__init__([local, remote])

    def start(self):
        if isinstance(self._remote, TaktukRemote):
            wait_all_taktuk_processes_start = _ChainPutCopyTaktukProcessLH(
                len(self._remote.processes),
                self._local)
            for p in self._remote.processes:
                p.lifecycle_handlers.append(wait_all_taktuk_processes_start)
            retval = super(ParallelActions, self).start()
            self._remote.start()
            return retval
        else:
            return super(_ChainPutCopy, self).start()

_execo_chainput = os.path.abspath(os.path.join(os.path.dirname(__file__), "execo-chainput"))
if not os.path.isfile(_execo_chainput): _execo_chainput = None

class ChainPut(SequentialActions):

    """Broadcast local files to several remote host, with an unencrypted, unauthenticated chain of host to host copies (idea taken from `kastafior <https://gforge.inria.fr/plugins/scmgit/cgi-bin/gitweb.cgi?p=kadeploy3/kadeploy3.git;a=tree;f=addons/kastafior;h=e5472ce9e800c80d9f54d1097ebbcba77f8ccd7a;hb=3.1.7>`_).

    Each broadcast is performed with a chain copy (simultaneously:
    host0 sending to host1, host1 sending to host2, ... hostN to
    hostN+1)

    ChainPut relies on:

    - running a bourne shell and netcat being available both on remote
      hosts and on localhost.

    - direct TCP connections allowed between any nodes among localhost
      and remote hosts. The exact chain of TCP connections is: localhost
      to first remote host, first remote host to second remote host, and
      so on up to the last remote host.

    On the security side, data transfers are not crypted, and ChainPut
    briefly opens a TCP server socket on each remote host, accepting
    any data without authentication. Insecure temporary files are
    used. It is thus intended to be used in a secured network
    environment.
    """

    def __init__(self, hosts, local_files, remote_location = ".", connection_params = None, **kwargs):
        """
        :param hosts: iterable of `execo.host.Host` onto which to copy
          the files.

        :param local_file: iterable of source file (local pathes).

        :param remote_location: destination directory (remote path).

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.
        """
        self.hosts = hosts
        # self.good_hosts = set(self.hosts)
        # self.bad_hosts = set()
        self.local_files = local_files
        self.remote_location = remote_location
        self.connection_params = connection_params
        if "name" not in kwargs:
            kwargs.update({"name": "%s to %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(ChainPut, self).__init__([], **kwargs)
        self.hide_subactions = True

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, v):
        self._hosts = get_unique_hosts_list(singleton_to_collection(v))

    def _init_actions(self):
        if len(self.hosts) > 0:
            actual_connection_params = make_connection_params(self.connection_params)
            chain_retries = actual_connection_params['chainput_chain_retry']
            if isinstance(chain_retries, float):
                chain_retries = int(chain_retries * len(self.hosts))

            chainhosts_handle, chainhosts_filename = tempfile.mkstemp(prefix = 'tmp_execo_chainhosts_')
            chainhosts = "\n".join([h.address for h in self.hosts]) + "\n"
            os.write(chainhosts_handle, chainhosts)
            os.close(chainhosts_handle)

            chainscript_filename = tempfile.mktemp(prefix = 'tmp_execo_chainscript_')
            if not _execo_chainput:
                raise EnvironmentError("unable to find execo-chainput")
            shutil.copy2(_execo_chainput, chainscript_filename)
            os.chmod(chainscript_filename, os.stat(chainscript_filename).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH )

            preparechain = TaktukPut(self.hosts,
                                     [ chainhosts_filename, chainscript_filename ],
                                     "/tmp",
                                     actual_connection_params
                                     )

            #previous_action = preparechain
            chains = []
            port = get_port()
            for findex, f in enumerate(self.local_files):

                autoremoveopt = ""
                if findex + 1 >= len(self.local_files):
                    autoremoveopt = " --autoremove"

                fwdcmd = [ "%s '%s' '%s' '%s' %i %i %i %i %i %i %i '%s'%s" % (
                        chainscript_filename,
                        f,
                        self.remote_location,
                        actual_connection_params['nc'],
                        actual_connection_params['chainput_nc_client_timeout'],
                        actual_connection_params['chainput_nc_server_timeout'],
                        port,
                        actual_connection_params['chainput_host_retry'],
                        chain_retries,
                        actual_connection_params['chainput_try_delay'],
                        idx+1,
                        chainhosts_filename,
                        autoremoveopt,
                        ) for idx, host in enumerate(self.hosts) ]

                fwd = TaktukRemote("{{fwdcmd}}",
                                   self.hosts,
                                   actual_connection_params)

                send = Local("%s '%s' '%s' '%s' %i %i %i %i %i %i %i '%s'%s" % (
                        chainscript_filename,
                        f,
                        self.remote_location,
                        actual_connection_params['nc'],
                        actual_connection_params['chainput_nc_client_timeout'],
                        actual_connection_params['chainput_nc_server_timeout'],
                        port,
                        actual_connection_params['chainput_host_retry'],
                        chain_retries,
                        actual_connection_params['chainput_try_delay'],
                        0,
                        chainhosts_filename,
                        autoremoveopt,
                        ))

                chain = _ChainPutCopy(send, fwd)
                #previous_action.lifecycle_handlers.insert(0, _ChainPutActionHostFilteringLH(self, [fwd]))
                #previous_action = chain
                chains.append(chain)

            self.actions = [ preparechain ] + chains
        else:
            self.actions = []
        super(ChainPut, self)._init_actions()

class RemoteSerial(Remote):

    """Open a serial port on several hosts in parallel through ``ssh`` or a similar remote connection tool.

    The serial port can be read (standard output) and written to (standard input).
    """

    def __init__(self, hosts, device, speed, connection_params = None, process_args = None, **kwargs):
        """:param hosts: iterable of `execo.host.Host` to which to
          connect and open the serial device.

        :param device: The Path to the serial device on the remote
          hosts (for example: ``/dev/ttyUSB1``). Substitions described
          in `execo.substitutions.remote_substitute` will be
          performed.

        :param speed: the speed of the serial port (for example:
          115200)

        :param connection_params: a dict similar to
          `execo.config.default_connection_params` whose values will
          override those in default_connection_params for connection.

        :param process_args: Dict of keyword arguments passed to
          instanciated processes.
        """
        self.hosts = hosts
        """Iterable of `execo.host.Host` to which to connect and run the command."""
        if "name" not in kwargs:
            kwargs.update({"name": "%s to %i hosts" % (self.__class__.__name__, len(self.hosts))})
        super(Remote, self).__init__(**kwargs)
        self.connection_params = connection_params
        """A dict similar to `execo.config.default_connection_params` whose values
        will override those in default_connection_params for connection."""
        self.device = device
        """Path to the serial devices on the remote hosts. (for example:
        ``/dev/ttyUSB1``). Substitions described in
        `execo.substitutions.remote_substitute` will be performed."""
        self.speed = speed
        """The speed of the serial port (for example: 115200)"""
        if process_args != None:
            self.process_args = process_args
            """Dict of keyword arguments passed to instanciated processes."""
        else:
            self.process_args = {}
        self._caller_context = get_caller_context()
        self._init_processes()

    def _args(self):
        return [ repr(self.hosts),
                 repr(self.device),
                 repr(self.speed) ] + Action._args(self) + RemoteSerial._kwargs(self)

    def _kwargs(self):
        kwargs = []
        if self.connection_params: kwargs.append("connection_params=%r" % (self.connection_params,))
        return kwargs

    def _init_processes(self):
        self.processes = []
        processlh = ActionNotificationProcessLH(self, len(self.hosts))
        for (index, host) in enumerate(self.hosts):
            p = SerialSsh(host,
                          remote_substitute(self.device, self.hosts, index, self._caller_context),
                          self.speed,
                          connection_params = self.connection_params,
                          **self.process_args)
            p.lifecycle_handlers.append(processlh)
            self.processes.append(p)

class ActionFactory:
    """Instanciate multiple remote process execution and file copies using configurable connector tools: ``ssh``, ``scp``, ``taktuk``"""

    def __init__(self, remote_tool = None, fileput_tool = None, fileget_tool = None):
        """
        :param remote_tool: can be `execo.config.SSH` or
          `execo.config.TAKTUK`

        :param fileput_tool: can be `execo.config.SCP`,
          `execo.config.TAKTUK` or `execo.config.CHAINPUT`

        :param fileget_tool: can be `execo.config.SCP` or
          `execo.config.TAKTUK`
        """

        if remote_tool == None:
            remote_tool = configuration.get("remote_tool")
        if fileput_tool == None:
            fileput_tool = configuration.get("fileput_tool")
        if fileget_tool == None:
            fileget_tool = configuration.get("fileget_tool")
        self.remote_tool = remote_tool
        self.fileput_tool = fileput_tool
        self.fileget_tool = fileget_tool

    def get_remote(self, *args, **kwargs):
        """Instanciates a `execo.action.Remote` or `execo.action.TaktukRemote`"""
        if self.remote_tool == SSH:
            return Remote(*args, **kwargs)
        elif self.remote_tool == TAKTUK:
            return TaktukRemote(*args, **kwargs)
        else:
            raise KeyError("no such remote tool: %s" % self.remote_tool)

    def get_fileput(self, *args, **kwargs):
        """Instanciates a `execo.action.Put`, `execo.action.TaktukPut` or `execo.action.ChainPut`"""
        if self.fileput_tool == SCP:
            return Put(*args, **kwargs)
        elif self.fileput_tool == TAKTUK:
            return TaktukPut(*args, **kwargs)
        elif self.fileput_tool == CHAINPUT:
            return ChainPut(*args, **kwargs)
        else:
            raise KeyError("no such fileput tool: %s" % self.fileput_tool)

    def get_fileget(self, *args, **kwargs):
        """Instanciates a `execo.action.Get` or `execo.action.TaktukGet`"""
        if self.fileget_tool == SCP:
            return Get(*args, **kwargs)
        elif self.fileget_tool == TAKTUK:
            return TaktukGet(*args, **kwargs)
        else:
            raise KeyError("no such fileget tool: %s" % self.fileget_tool)

default_action_factory = ActionFactory()

def get_remote(*args, **kwargs):
    """Instanciates a `execo.action.Remote` or `execo.action.TaktukRemote` with the default factory"""
    return default_action_factory.get_remote(*args, **kwargs)

def get_fileput(*args, **kwargs):
    """Instanciates a `execo.action.Put` or `execo.action.TaktukPut` with the default factory"""
    return default_action_factory.get_fileput(*args, **kwargs)

def get_fileget(*args, **kwargs):
    """Instanciates a `execo.action.Get` or `execo.action.TaktukGet` with the default factory"""
    return default_action_factory.get_fileget(*args, **kwargs)



# Copyright 2009-2023 INRIA Rhone-Alpes, Service Experimentation et
# Developpement
#
# This file is part of Execo.
#
# Execo is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Execo is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Execo.  If not, see <http://www.gnu.org/licenses/>

from .config import g5k_configuration
from execo.action import Remote, ActionNotificationProcessLH, \
    Action, get_remote
from execo.config import make_connection_params
from execo.host import get_hosts_set, Host
from execo.log import style, logger
from execo.process import ProcessOutputHandler, get_process, \
    handle_process_output
from execo.time_utils import format_seconds
from execo.utils import comma_join, compact_output, singleton_to_collection
from execo_g5k.config import default_frontend_connection_params
from execo_g5k.utils import get_frontend_host, get_kavlan_host_name
from .utils import get_default_frontend
import copy
import re
import time

class Deployment(object):
    """A kadeploy3 deployment, POD style class."""

    def __init__(self,
                 hosts = None,
                 env_file = None,
                 env_name = None,
                 user = None,
                 vlan = None,
                 other_options = None):
        self.hosts = hosts
        """hosts: iterable of `execo.host.Host` on which to deploy."""
        self.env_file = env_file
        """filename of an environment to deploy"""
        self.env_name = env_name
        """name of a kadeploy3 registered environment to deploy.

        there must be either one of env_name or env_file parameter
        given. If none given, will try to use the default environement
        from `g5k_configuration`."""
        self.user = user
        """kadeploy3 user"""
        self.vlan = vlan
        """if given, kadeploy3 will automatically switch the nodes to this vlan after deployment"""
        self.other_options = other_options
        """string of other options to pass to kadeploy3"""

    def _get_common_kadeploy_command_line(self):
        cmd_line = g5k_configuration.get('kadeploy3')
        cmd_line += " " + g5k_configuration.get('kadeploy3_options')
        if self.env_file and self.env_name:
            raise ValueError("Deployment cannot have both env_file and env_name")
        if (not self.env_file) and (not self.env_name):
            if g5k_configuration.get('default_env_name') and g5k_configuration.get('default_env_file'):
                raise Exception("g5k_configuration cannot have both default_env_name and default_env_file")
            if (not g5k_configuration.get('default_env_name')) and (not g5k_configuration.get('default_env_file')):
                raise Exception("no environment name or file found")
            if g5k_configuration.get('default_env_name'):
                cmd_line += " -e %s" % (g5k_configuration['default_env_name'],)
            elif g5k_configuration.get('default_env_file'):
                cmd_line += " -a %s" % (g5k_configuration['default_env_file'],)
        elif self.env_name:
            cmd_line += " -e %s" % (self.env_name,)
        elif self.env_file:
            cmd_line += " -a %s" % (self.env_file,)
        if self.user != None:
            cmd_line += " -u %s" % (self.user,)
        if self.vlan != None:
            cmd_line += " --vlan %s" % (self.vlan,)
        if self.other_options:
            cmd_line += " %s" % (self.other_options,)
        return cmd_line

    def __repr__(self):
        s = ""
        if self.hosts != None: s = comma_join(s, "hosts=%r" % (self.hosts,))
        if self.env_file != None: s = comma_join(s, "env_file=%r" % (self.env_file,))
        if self.env_name != None: s = comma_join(s, "env_name=%r" % (self.env_name,))
        if self.user != None: s = comma_join(s, "user=%r" % (self.user,))
        if self.vlan != None: s = comma_join(s, "vlan=%r" % (self.vlan,))
        if self.other_options: s = comma_join(s, "other_options=%r" % (self.other_options,))
        return "Deployment(%s)" % (s,)

_ksoh_deployed_nodes_header_re1 = re.compile("^Nodes correctly deployed on cluster \w+\s*$")       # for kadeploy3 < 3.2
_ksoh_undeployed_nodes_header_re1 = re.compile("^Nodes not correctly deployed on cluster \w+\s*$") # for kadeploy3 < 3.2
_ksoh_deployed_nodes_header_re2 = re.compile("^The \w+ is successful on nodes\s*$")                # for kadeploy3 >= 3.2
_ksoh_undeployed_nodes_header_re2 = re.compile("^The \w+ failed on nodes\s*$")                     # for kadeploy3 >= 3.2
_ksoh_deployed_node_re = re.compile("^(\w+-\d+\.\w+\.grid5000\.fr)\s*$")
_ksoh_undeployed_node_re = re.compile("^(\w+-\d+\.\w+\.grid5000\.fr)(\s+\(.*)?\s*$")

class _KadeployStdoutHandler(ProcessOutputHandler):

    """Parse kadeploy3 stdout."""

    def __init__(self):
        super(_KadeployStdoutHandler, self).__init__()
        self._SECTION_NONE, self._SECTION_DEPLOYED_NODES, self._SECTION_UNDEPLOYED_NODES = list(range(3))
        self._current_section = self._SECTION_NONE

    def action_reset(self):
        self._current_section = self._SECTION_NONE

    def read_line(self, process, stream, string):
        if (_ksoh_deployed_nodes_header_re1.search(string) != None
            or _ksoh_deployed_nodes_header_re2.search(string) != None):
            self._current_section = self._SECTION_DEPLOYED_NODES
            return
        if (_ksoh_undeployed_nodes_header_re1.search(string) != None
            or _ksoh_undeployed_nodes_header_re2.search(string) != None):
            self._current_section = self._SECTION_UNDEPLOYED_NODES
            return
        if self._current_section == self._SECTION_DEPLOYED_NODES:
            so = _ksoh_deployed_node_re.search(string)
            if so != None:
                host_address = so.group(1)
                process.kadeployer.deployed_hosts.add(host_address)
                process.deployed_hosts.add(host_address)
        elif self._current_section == self._SECTION_UNDEPLOYED_NODES:
            so = _ksoh_undeployed_node_re.search(string)
            if so != None:
                host_address = so.group(1)
                process.kadeployer.undeployed_hosts.add(host_address)
                process.undeployed_hosts.add(host_address)

_host_site_re1 = re.compile("^[^ \t\n\r\f\v\.]+\.([^ \t\n\r\f\v\.]+)\.grid5000.fr$")
_host_site_re2 = re.compile("^[^ \t\n\r\f\v\.]+\.([^ \t\n\r\f\v\.]+)$")
_host_site_re3 = re.compile("^[^ \t\n\r\f\v\.]+$")

def _get_host_frontend(host):
    # Get the frontend of a host.
    #
    # This function is specific to kadeploy because:
    #
    # - only handles execo.Host
    #
    # - we could use get_host_site but api_utils is not always
    #   available (if api is down or if requests is not available)
    #   -> this is not true anymore (13/2/2014)
    #
    # - special bahavior: fallback to default frontend if unable to
    #   extract site from host name
    #
    # - raises exception if host name format invalid (not sure we
    #   would want this for a generic function)
    frontend = None
    mo1 = _host_site_re1.search(host.address)
    if mo1 != None:
        frontend = mo1.group(1)
    else:
        mo2 = _host_site_re2.search(host.address)
        if mo2 != None:
            frontend = mo1.group(1)
        else:
            mo3 = _host_site_re3.search(host.address)
            if mo3 != None:
                frontend = get_default_frontend()
            else:
                raise ValueError("unknown frontend for host %s" % host.address)
    return frontend

class FrontendPrefixWrapper(ProcessOutputHandler):

    def __init__(self, handler):
        super(FrontendPrefixWrapper, self).__init__()
        self.handler = handler

    def read_line(self, process, stream, string):
        handle_process_output(process,
                              stream,
                              self.handler,
                              process.frontend + ": " + string,
                              False, False)

class Kadeployer(Remote):

    """Deploy an environment with kadeploy3 on several hosts.

    Able to deploy in parallel to multiple frontends.
    """

    def __init__(self, deployment, frontend_connection_params = None,
                 stdout_handlers = None, stderr_handlers = None):
        """
        :param deployment: instance of Deployment class describing the
          intended kadeployment.

        :param frontend_connection_params: connection params for
          connecting to frontends if needed. Values override those in
          `execo_g5k.config.default_frontend_connection_params`.

        :param stdout_handlers: iterable of `ProcessOutputHandlers`
          which will be passed to the actual deploy processes.

        :param stderr_handlers: iterable of `ProcessOutputHandlers`
          which will be passed to the actual deploy processes.
        """
        super(Remote, self).__init__()
        self.deployed_hosts = set()
        """Iterable of `Host` containing the deployed hosts. This iterable won't be
        complete if the Kadeployer has not terminated."""
        self.undeployed_hosts = set()
        """Iterable of `Host` containing the hosts not deployed. This iterable
        won't be complete if the Kadeployer has not terminated."""
        self._stdout_handlers = stdout_handlers
        self._stderr_handlers = stderr_handlers
        """Iterable of `ProcessOutputHandlers` which will be passed to the
        actual deploy processes."""
        self.frontend_connection_params = frontend_connection_params
        """Connection params for connecting to frontends if needed. Values override
        those in `execo_g5k.config.default_frontend_connection_params`."""
        self.deployment = deployment
        """Instance of Deployment class describing the intended kadeployment."""
        self.timeout = None
        """Deployment timeout"""
        self._init_processes()
        self.name = "%s on %i hosts / %i frontends" % (self.__class__.__name__, len(self._unique_hosts), len(self.processes))

    def _init_processes(self):
        self.processes = []
        self._unique_hosts = get_hosts_set(self.deployment.hosts)
        frontends = dict()
        for host in self._unique_hosts:
            frontend = _get_host_frontend(host)
            if frontend in frontends:
                frontends[frontend].append(host)
            else:
                frontends[frontend] = [host]
        lifecycle_handler = ActionNotificationProcessLH(self, len(frontends))
        deploy_stdout_handler = _KadeployStdoutHandler()
        for frontend in frontends:
            kadeploy_command = self.deployment._get_common_kadeploy_command_line()
            for host in frontends[frontend]:
                kadeploy_command += " -m %s" % (host.address,)
            p = get_process(kadeploy_command,
                            host = get_frontend_host(frontend),
                            connection_params = make_connection_params(self.frontend_connection_params,
                                                                     default_frontend_connection_params))
            p.pty = True
            p.timeout = self.timeout
            p.stdout_handlers.append(deploy_stdout_handler)
            p.stdout_handlers.extend([ FrontendPrefixWrapper(h)
                                       for h in singleton_to_collection(self._stdout_handlers) ])
            p.stderr_handlers.extend([ FrontendPrefixWrapper(h)
                                       for h in singleton_to_collection(self._stderr_handlers) ])
            p.lifecycle_handlers.append(lifecycle_handler)
            p.frontend = frontend
            p.kadeploy_hosts = [ host.address for host in frontends[frontend] ]
            p.deployed_hosts = set()
            p.undeployed_hosts = set()
            p.kadeployer = self
            self.processes.append(p)

    def _common_reset(self):
        super(Kadeployer, self)._common_reset()
        self.deployed_hosts = set()
        self.undeployed_hosts = set()

    def _check_ok(self, log):
        ok = True
        error_logs = []
        warn_logs = []
        for process in self.processes:
            if ( len(process.deployed_hosts.intersection(process.undeployed_hosts)) != 0
                 or len(process.deployed_hosts.union(process.undeployed_hosts).symmetric_difference(process.kadeploy_hosts)) != 0 ):
                error_logs.append("deploy on %s, total/deployed/undeployed = %i/%i/%i:\n%s\nstdout:\n%s\nstderr:\n%s" % (
                        process.frontend, len(process.kadeploy_hosts), len(process.deployed_hosts), len(process.undeployed_hosts),
                        process, compact_output(process.stdout), compact_output(process.stderr)))
                ok = False
            else:
                if len(process.deployed_hosts) == 0:
                    warn_logs.append("deploy on %s, total/deployed/undeployed = %i/%i/%i:\n%s\nstdout:\n%s\nstderr:\n%s" % (
                            process.frontend, len(process.kadeploy_hosts), len(process.deployed_hosts), len(process.undeployed_hosts),
                            process, compact_output(process.stdout), compact_output(process.stderr)))
                    ok = False
                elif len(process.undeployed_hosts) > 0:
                    warn_logs.append("deploy on %s, total/deployed/undeployed = %i/%i/%i" % (
                            process.frontend, len(process.kadeploy_hosts), len(process.deployed_hosts), len(process.undeployed_hosts)))
        if log:
            if len(warn_logs) > 0:
                logger.warn(str(self) + ":\n" + "\n".join(warn_logs))
            if len(error_logs) > 0:
                logger.error(str(self) + ":\n" + "\n".join(error_logs))
        return ok

    def _notify_terminated(self):
        self._check_ok(True)
        super(Kadeployer, self)._notify_terminated()

    def _args(self):
        return [ repr(self.deployment) ] + Action._args(self) + Kadeployer._kwargs(self)

    def _kwargs(self):
        kwargs = []
        if self.frontend_connection_params: kwargs.append("frontend_connection_params=%r" % (self.frontend_connection_params,))
        return kwargs

    def _infos(self):
        return super(Remote, self)._infos() + [ "total/deployed/undeployed = %i/%i/%i" % (
                len(self._unique_hosts), len(self.deployed_hosts), len(self.undeployed_hosts),) ]

    @property
    def ok(self):
        return (not self.ended or self._check_ok(False)) and super(Kadeployer, self).ok

def deploy(deployment,
           check_deployed_command = True,
           node_connection_params = {'user': 'root'},
           num_tries = 1,
           check_enough_func = None,
           frontend_connection_params = None,
           deploy_timeout = None,
           check_timeout = 30,
           stdout_handlers = None,
           stderr_handlers = None):
    """Deploy nodes, many times if needed, checking which of these nodes are already deployed with a user-supplied command. If no command given for checking if nodes deployed, rely on kadeploy to know which nodes are deployed.

    - loop `num_tries` times:

      - if ``check_deployed_command`` given, try to connect to these
        hosts using the supplied `node_connection_params` (or the
        default ones), and to execute ``check_deployed_command``. If
        connection succeeds and the command returns 0, the host is
        assumed to be deployed, else it is assumed to be undeployed.

      - optionnaly call user-supplied ``check_enough_func``, passing
        to it the list of deployed and undeployed hosts, to let user
        code decide if enough nodes deployed. Otherwise, try as long
        as there are undeployed nodes.

      - deploy the undeployed nodes

    returns a tuple with the list of deployed hosts and the list of
    undeployed hosts.

    When checking correctly deployed nodes with
    ``check_deployed_command``, and if the deployment is using the
    kavlan option, this function will try to contact the nodes using
    the appropriate DNS hostnames in the new vlan.

    :param deployment: instance of `execo.kadeploy.Deployment` class
      describing the intended kadeployment.

    :param check_deployed_command: command to perform remotely to
      check node deployement. May be a String, True, False or None. If
      String: the actual command to be used (This command should
      return 0 if the node is correctly deployed, or another value
      otherwise). If True, the default command value will be used
      (from `execo_g5k.config.g5k_configuration`). If None or False,
      no check is made and deployed/undeployed status will be taken
      from kadeploy's output.

    :param node_connection_params: a dict similar to
      `execo.config.default_connection_params` whose values will
      override those in `execo.config.default_connection_params` when
      connecting to check node deployment with
      ``check_deployed_command`` (see below).

    :param num_tries: number of deploy tries

    :param check_enough_func: a function taking as parameter a list of
      deployed hosts and a list of undeployed hosts, which will be
      called at each deployment iteration end, and that should return
      a boolean indicating if there is already enough nodes (in this
      case, no further deployement will be attempted).

    :param frontend_connection_params: connection params for connecting
      to frontends if needed. Values override those in
      `execo_g5k.config.default_frontend_connection_params`.

    :param deploy_timeout: timeout for deployement. Default is None,
      which means no timeout.

    :param check_timeout: timeout for node deployment checks. Default
      is 30 seconds.

    :param stdout_handlers: iterable of `ProcessOutputHandlers`
          which will be passed to the actual deploy processes.

    :param stderr_handlers: iterable of `ProcessOutputHandlers`
          which will be passed to the actual deploy processes.
    """

    if check_enough_func == None:
        check_enough_func = lambda deployed, undeployed: len(undeployed) == 0

    if check_deployed_command == True:
        check_deployed_command = g5k_configuration.get('check_deployed_command')

    def check_update_deployed(undeployed_hosts, check_deployed_command, node_connection_params, vlan): #IGNORE:W0613
        logger.debug(style.emph("check which hosts are already deployed among:") + " %s", undeployed_hosts)
        deployment_hostnames_mapping = dict()
        if vlan:
            for host in undeployed_hosts:
                deployment_hostnames_mapping[get_kavlan_host_name(host, vlan)] = host
        else:
            for host in undeployed_hosts:
                deployment_hostnames_mapping[host] = host
        deployed_check = get_remote(check_deployed_command,
                                    list(deployment_hostnames_mapping),
                                    connection_params = node_connection_params)
        for p in deployed_check.processes:
                p.nolog_exit_code = True
                p.nolog_timeout = True
                p.nolog_error = True
                p.timeout = check_timeout
        deployed_check.run()
        newly_deployed = list()
        for process in deployed_check.processes:
            logger.debug(style.emph("check on %s:" % (process.host,))
                         + " %s\n" % (process,)
                         + style.emph("stdout:") + "\n%s\n" % (process.stdout)
                         + style.emph("stderr:") + "\n%s\n" % (process.stderr))
            if (process.ok):
                newly_deployed.append(deployment_hostnames_mapping[process.host.address])
                logger.debug("OK %s", deployment_hostnames_mapping[process.host.address])
            else:
                logger.debug("KO %s", deployment_hostnames_mapping[process.host.address])
        return newly_deployed

    start_time = time.time()
    deployed_hosts = set()
    undeployed_hosts = set([ Host(host).address for host in deployment.hosts ])
    my_newly_deployed = []
    if check_deployed_command:
        my_newly_deployed = check_update_deployed(undeployed_hosts, check_deployed_command, node_connection_params, deployment.vlan)
        deployed_hosts.update(my_newly_deployed)
        undeployed_hosts.difference_update(my_newly_deployed)
    num_tries_done = 0
    elapsed = time.time() - start_time
    last_time = time.time()
    deploy_stats = list()   # contains tuples ( timestamp,
                            #                   num attempted deploys,
                            #                   len(kadeployer.deployed_hosts),
                            #                   len(my_newly_deployed),
                            #                   len(deployed_hosts),
                            #                   len(undeployed_hosts )
    deploy_stats.append((elapsed, None, None, len(my_newly_deployed), len(deployed_hosts), len(undeployed_hosts)))
    while (not check_enough_func(deployed_hosts, undeployed_hosts)
           and num_tries_done < num_tries):
        num_tries_done += 1
        logger.debug(style.emph("try %i, deploying on:" % (num_tries_done,)) + " %s", undeployed_hosts)
        tmp_deployment = copy.copy(deployment)
        tmp_deployment.hosts = undeployed_hosts
        kadeployer = Kadeployer(tmp_deployment,
                                frontend_connection_params = frontend_connection_params,
                                stdout_handlers = stdout_handlers,
                                stderr_handlers = stderr_handlers)
        kadeployer.timeout = deploy_timeout
        kadeployer.run()
        my_newly_deployed = []
        if check_deployed_command:
            my_newly_deployed = check_update_deployed(undeployed_hosts, check_deployed_command, node_connection_params, deployment.vlan)
            deployed_hosts.update(my_newly_deployed)
            undeployed_hosts.difference_update(my_newly_deployed)
        else:
            deployed_hosts.update(kadeployer.deployed_hosts)
            undeployed_hosts.difference_update(kadeployer.deployed_hosts)
        logger.debug(style.emph("kadeploy reported newly deployed hosts:") + "   %s", kadeployer.deployed_hosts)
        logger.debug(style.emph("check reported newly deployed hosts:") + "   %s", my_newly_deployed)
        logger.debug(style.emph("all deployed hosts:") + "     %s", deployed_hosts)
        logger.debug(style.emph("still undeployed hosts:") + " %s", undeployed_hosts)
        elapsed = time.time() - last_time
        last_time = time.time()
        deploy_stats.append((elapsed,
                             len(tmp_deployment.hosts),
                             len(kadeployer.deployed_hosts),
                             len(my_newly_deployed),
                             len(deployed_hosts),
                             len(undeployed_hosts)))

    logger.detail(style.emph("deploy finished") + " in %i tries, %s", num_tries_done, format_seconds(time.time() - start_time))
    logger.detail("deploy  duration  attempted  deployed     deployed     total     total")
    logger.detail("                  deploys    as reported  as reported  already   still")
    logger.detail("                             by kadeploy  by check     deployed  undeployed")
    logger.detail("---------------------------------------------------------------------------")
    for (deploy_index, deploy_stat) in enumerate(deploy_stats):
        logger.detail("#%-5.5s  %-8.8s  %-9.9s  %-11.11s  %-11.11s  %-8.8s  %-10.10s",
            deploy_index,
            format_seconds(deploy_stat[0]),
            deploy_stat[1],
            deploy_stat[2],
            deploy_stat[3],
            deploy_stat[4],
            deploy_stat[5])
    logger.debug(style.emph("deployed hosts:") + " %s", deployed_hosts)
    logger.debug(style.emph("undeployed hosts:") + " %s", undeployed_hosts)

    return (deployed_hosts, undeployed_hosts)
